from __future__ import print_function, division
from PyAstronomy.pyaC import pyaPermanent as pp
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy import pyaC
from PyAstronomy.pyasl import _ic
import os
import gzip
import csv
import numpy as np
import warnings
import six.moves as smo
import six
import copy

if _ic.check["astropy"]:
    from astropy.io import votable


class ExoplanetEU(pp.PyAUpdateCycle):
    """
    Provides access to exoplanet.eu data base.

    This class downloads the data base as a csv
    file and converts it into a numpy recarray.
    By default, the data are re-downloaded
    every 7 days.

    The available columns are:

    ============  ==============================  ======
     Column Name                     Description    Unit

          plName                  Name of planet
          plMass                  Mass of planet      MJ
        plRadius                Radius of planet      RJ
          period                  Orbital period       d
             sma                 Semi-major axis      AU
    eccentricity            Orbital eccentricity
     inclination             Orbital inclination     deg
     angDistance                Angular Distance  arcsec
       pubStatus              Publication status
      discovered               Year of discovery      yr
         updated             Date of data update
           omega          Argument of Periastron     deg
           tperi             Epoch of Periastron       d
         detType                  Detection type
       molecules      List of detected molecules
          stName                    Name of star
              ra         Right ascension (J2000)     hms
             dec             Declination (J2000)     dms
           mag_v      V magnitude of a host star     mag
           mag_i      I magnitude of a host star     mag
           mag_j      J magnitude of a host star     mag
           mag_h      H magnitude of a host star     mag
           mag_k      K magnitude of a host star     mag
            dist           Distance to host star      pc
              mh        Metallicity of host star     dex
          stMass                    Stellar mass   solar
        stRadius                  Radius of star   solar
             SpT      Spectral type of host star
           stAge                     Stellar age      Ga
          stTeff   Stellar effective temperature       K
        plRadMM          Measuring method of Rpl
    ============  ==============================  ======

    Parameters
    ----------
    skipUpdate : boolean, optional
        If True, no re-download of the data
        will be initiated no matter how old
        they are.

    """

    def _download(self):
        """
        Download data.
        """
        try:
            self._fs.downloadToFile(
                "http://exoplanet.eu/catalog/csv",
                self.dataFileName,
                clobber=True,
                verbose=False,
                openMethod=gzip.open,
            )
        except PE.PyADownloadError as pde:
            pde.addInfo = "Unfortunately, this is a known bug for which no fix could so far be provided.\n"
            pde.addInfo += "    Please try ExoplanetEU2 instead."
            raise

    def _readData(self):
        """ """
        # Determine number of planets in the csv file
        r = csv.DictReader(
            self._fs.requestFile(self.dataFileName, "rt", gzip.open), delimiter=","
        )
        for nplanets, x in enumerate(r):
            pass
        # Reinitialize csv file
        r = csv.DictReader(
            self._fs.requestFile(self.dataFileName, "rt", gzip.open), delimiter=","
        )
        # Determine data types for numpy recarray from columns
        # and initialize
        dtype = [
            (self._columns[x][0], self._columns[x][3])
            for x in range(len(self._columns))
        ]
        self.data = np.recarray((nplanets + 1,), dtype=dtype)
        colnotfilled = [self._columns[x][0] for x in six.iterkeys(self._columns)]
        for i, x in enumerate(r):
            for k, v in six.iteritems(x):
                # Remove hash and white spaces from column names
                k = k.strip("#")
                k = k.strip()
                # Translate csv column name into internal column name
                if k in self._ident:
                    key = self._ident[k]
                else:
                    key = k
                if len(v) == 0:
                    v = None
                # Accept only expected fields
                if not key in self.data.dtype.names:
                    continue
                try:
                    colnotfilled.remove(key)
                except ValueError:
                    # Ignoring already removed value
                    pass
                self.data[key][i] = v
        if len(colnotfilled) > 0:
            PE.warn(
                PE.PyAAlgorithmFailure(
                    "Not all columns could be filled with data. The following columns must not be used: "
                    + ", ".join(colnotfilled),
                    where="ExoplanetEU",
                    solution="The format of the data base must be checked. Please consider issuing a bug report via github.",
                )
            )

    def availableColumns(self):
        """
        Show a summary of the available columns.

        Returns
        -------
        Column names : list of strings
            The names of the columns.
        """
        print("-" * 51)
        print("%12s  %30s  %5s" % ("Column Name", "Description", "Unit"))
        print("-" * 51)
        cols = []
        for k, v in six.iteritems(self._columns):
            print("%12s  %30s  %5s" % tuple(v[0:3]))
            cols.append(v[0])
        print("-" * 51)
        return cols

    def getAllData(self):
        """
        Provides all data as a numpy recarray.

        Returns
        -------
        Data : numpy recarray
            All available data. Use, e.g.,
            `availableColumns` to get an
            overview of the available
            information.
        """
        return self.data.copy()

    def forceUpdate(self):
        """
        Force a fresh download of the data.

        By default, the data will be updated every
        7 days.
        """
        self._update(self._download)

    def __init__(self, skipUpdate=False):
        configFilename = os.path.join("pyasl", "resBased", "epeu.cfg")
        pp.PyAUpdateCycle.__init__(self, configFilename, "ExoUpdate")
        self.dataFileName = os.path.join("pyasl", "resBased", "epeu.csv.gz")
        self._fs = pp.PyAFS()
        if self.needsUpdate() and (not skipUpdate):
            self._update(self._download)

        # Define internal column names and data types
        self._columns = {}
        self._columns[0] = ["plName", "Name of planet", "", "U15"]
        self._columns[1] = ["plMass", "Mass of planet", "MJ", float]
        self._columns[2] = ["plRadius", "Radius of planet", "RJ", float]
        self._columns[3] = ["period", "Orbital period", "d", float]
        self._columns[4] = ["sma", "Semi-major axis", "AU", float]
        self._columns[5] = ["eccentricity", "Orbital eccentricity", "", float]
        self._columns[6] = ["inclination", "Orbital inclination", "deg", float]
        self._columns[7] = ["angDistance", "Angular Distance", "arcsec", float]
        self._columns[8] = ["pubStatus", "Publication status", "", "U2"]
        self._columns[9] = ["discovered", "Year of discovery", "yr", float]
        self._columns[10] = ["updated", "Date of data update", "", "U10"]
        self._columns[11] = ["omega", "Argument of Periastron", "deg", float]
        self._columns[12] = ["tperi", "Epoch of Periastron", "d", float]
        self._columns[13] = ["detType", "Detection type", "", "U2"]
        self._columns[14] = ["molecules", "List of detected molecules", "", "U10"]
        self._columns[15] = ["stName", "Name of star", "", "U15"]
        self._columns[16] = ["ra", "Right ascension (J2000)", "hms", "U12"]
        self._columns[17] = ["dec", "Declination (J2000)", "dms", "U12"]
        self._columns[18] = ["mag_v", "V magnitude of a host star", "mag", float]
        self._columns[19] = ["mag_i", "I magnitude of a host star", "mag", float]
        self._columns[20] = ["mag_j", "J magnitude of a host star", "mag", float]
        self._columns[21] = ["mag_h", "H magnitude of a host star", "mag", float]
        self._columns[22] = ["mag_k", "K magnitude of a host star", "mag", float]
        self._columns[23] = ["dist", "Distance to host star", "pc", float]
        self._columns[24] = ["mh", "Metallicity of host star", "dex", float]
        self._columns[25] = ["stMass", "Stellar mass", "solar", float]
        self._columns[26] = ["stRadius", "Radius of star", "solar", float]
        self._columns[27] = ["SpT", "Spectral type of host star", "", "U5"]
        self._columns[28] = ["stAge", "Stellar age", "Ga", float]
        self._columns[29] = ["stTeff", "Stellar effective temperature", "K", float]
        self._columns[30] = ["plRadMM", "Measuring method of Rpl", "", "U15"]
        # Identify exoplanet.eu csv column names with internal column names
        self._ident = {
            "name": "plName",
            "mass": "plMass",
            "radius": "plRadius",
            "semi_major_axis": "sma",
            "angular_distance": "angDistance",
            "publication_status": "pubStatus",
            "detection_type": "detType",
            "star_name": "stName",
            "mag_v": "mag_v",
            "mag_i": "mag_i",
            "mag_j": "mag_j",
            "mag_h": "mag_h",
            "mag_k": "mag_k",
            "star_distance": "dist",
            "star_metallicity": "mh",
            "star_mass": "stMass",
            "star_radius": "stRadius",
            "star_sp_type": "SpT",
            "star_age": "stAge",
            "star_teff": "stTeff",
            "orbital_period": "period",
            "radius_detection_type": "plRadMM",
        }

        self._readData()


class ExoplanetEU2(pp.PyAUpdateCycle):
    """
    Provides access to exoplanet.eu data base.

    This class downloads the data base as a VO table.
    By default, the data are re-downloaded
    every 7 days.

    The class provides access to the entire data base
    offered by exoplanet.eu as shown below. Use `showAvailableData`
    to view the table of available data, including potential
    updates.

    ==========================  =======  =======  ============================================================================================
                          name    dtype     unit  description
    ==========================  =======  =======  ============================================================================================
                          name   object           Name of a planet
                          mass  float64  jovMass  Planetary Mass
                  mass_err_min  float64  jovMass  Planetary Mass error
                  mass_err_max  float64  jovMass  Planetary Mass error
                     mass_sini  float64  jovMass  Planetary Mass*sin(i)
           mass_sini_error_min  float64  jovMass  Planetary Mass*sin(i) error
           mass_sini_error_max  float64  jovMass  Planetary Mass*sin(i) error
                        radius  float64     Rjup  Planetary Radius
              radius_error_min  float64     Rjup  Planetary Radius error
              radius_error_max  float64     Rjup  Planetary Radius error
                orbital_period  float64        d  Orbital Period
        orbital_period_err_min  float64        d  Orbital Period error
        orbital_period_err_max  float64        d  Orbital Period error
               semi_major_axis  float64       AU  Semi-Major Axis
     semi_major_axis_error_min  float64       AU  Semi-Major Axis error
     semi_major_axis_error_max  float64       AU  Semi-Major Axis error
                  eccentricity  float64           Orbital Eccentricity
        eccentricity_error_min  float64           Orbital Eccentricity error
        eccentricity_error_max  float64           Orbital Eccentricity error
                   inclination  float64      deg  Orbital Inclination
         inclination_error_min  float64      deg  Orbital Inclination error
         inclination_error_max  float64      deg  Orbital Inclination error
              angular_distance  float64     arcs  Angular Distance
                    discovered    int32       yr  Year of Discovery
                       updated   object           Last Update
                         omega  float64      deg  Argument of Periastron
               omega_error_min  float64      deg  Argument of Periastron error
               omega_error_max  float64      deg  Argument of Periastron error
                         tperi  float64        d  Epoch of Periastron
               tperi_error_min  float64        d  Epoch of Periastron error
               tperi_error_max  float64        d  Epoch of Periastron error
                         tconj  float64        d  Conjonction Date
               tconj_error_min  float64        d  Conjonction Date error
               tconj_error_max  float64        d  Conjonction Date error
                      tzero_tr  float64        d  Primary Transit
            tzero_tr_error_min  float64        d  Primary Transit error
            tzero_tr_error_max  float64        d  Primary Transit error
                  tzero_tr_sec  float64        d  Secondary Transit
        tzero_tr_sec_error_min  float64        d  Secondary transit error
        tzero_tr_sec_error_max  float64        d  Secondary transit error
                  lambda_angle  float64      deg  Sky-projected angle between the planetary orbital spin and the stellar rotational spin
        lambda_angle_error_min  float64      deg  Sky-projected angle between the planetary orbital spin and the stellar rotational spin error
        lambda_angle_error_max  float64      deg  Sky-projected angle between the planetary orbital spin and the stellar rotational spin error
              impact_parameter  float64           Impact Parameter b
    impact_parameter_error_min  float64           Impact Parameter b error
    impact_parameter_error_max  float64           Impact Parameter b error
                      tzero_vr  float64        d  Zero Radial Speed time
            tzero_vr_error_min  float64        d  Zero Radial Speed time error
            tzero_vr_error_max  float64        d  Zero Radial Speed time error
                             k  float64    m / s  Velocity Semiamplitude K
                   k_error_min  float64    m / s  Velocity Semiamplitude K error
                   k_error_max  float64    m / s  Velocity Semiamplitude K error
               temp_calculated  float64        K  Calculated temperature
                 temp_measured  float64        K  Measured temperature
                 hot_point_lon  float64      deg  Hottest point longitude
              geometric_albedo  float64           Geometric albedo
    geometric_albedo_error_min  float64           Geometric albedo error
    geometric_albedo_error_max  float64           Geometric albedo error
                         log_g  float64           log(g)
            publication_status   object           Publication Status
                detection_type   object           Detection type
           mass_detection_type   object           Mass Measurement Method
         radius_detection_type   object           Radius Measurement Method
               alternate_names   object           List of planet alternative names
                     molecules   object           List of detected molecules
                     star_name   object           Name of a host star
                            ra  float64      deg  RA (J2000) of a star
                           dec  float64      deg  Dec (J2000) of a star
                         mag_v  float64      mag  V magnitude of a host star
                         mag_i  float64      mag  I magnitude of a host star
                         mag_j  float64      mag  J magnitude of a host star
                         mag_h  float64      mag  H magnitude of a host star
                         mag_k  float64      mag  K magnitude of a host star
                 star_distance  float64       pc  Distance to a host star
       star_distance_error_min  float64       pc  Distance to a host star error
       star_distance_error_max  float64       pc  Distance to a host star error
              star_metallicity  float64           Metallicity of a host star
    star_metallicity_error_min  float64           Metallicity of a host star error
    star_metallicity_error_max  float64           Metallicity of a host star error
                     star_mass  float64     Msun  Mass of a host star
           star_mass_error_min  float64     Msun  Mass of a host star error
           star_mass_error_max  float64     Msun  Mass of a host star error
                   star_radius  float64     Rsun  Radius of a host star
         star_radius_error_min  float64     Rsun  Radius of a host star error
         star_radius_error_max  float64     Rsun  Radius of a host star error
                  star_sp_type   object           Spectral type of a host star
                      star_age  float64      Gyr  Age of a host star
            star_age_error_min  float64      Gyr  Age of a host star error
            star_age_error_max  float64      Gyr  Age of a host star error
                     star_teff  float64        K  Effective temperature of a host star
           star_teff_error_min  float64        K  Effective temperature of a host star error
           star_teff_error_max  float64        K  Effective temperature of a host star error
            star_detected_disc   object           Star Detected Disc
           star_magnetic_field     bool           Star magnetic field
          star_alternate_names   object           List of star alternative names
    ==========================  =======  =======  ============================================================================================


    Parameters
    ----------
    skipUpdate : Boolean, optional
        If True, update of data will be skipped (default is False).
    forceUpdate : Boolean, optional
        If True, udpate of data will be forced (independent of value
        if `skipUpdate`, default is False)
    """

    def _readData(self):
        """
        Read data from VO table
        """
        # Temporarily suppress warnings (astropy issues one on reading this table)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                self.vot = votable.parse(
                    self._fs.requestFile(self.dataFileName, "r", gzip.open),
                    pedantic=False,
                    invalid="mask",
                )
            except Exception as e:
                votable.validate(
                    self._fs.requestFile(self.dataFileName, "r", gzip.open)
                )
                raise e
        # Use 'name' over ID field to specify column names
        self.vot = self.vot.get_first_table().to_table(use_names_over_ids=True)

    def forceUpdate(self):
        """
        Force a fresh download of the data and read the data.

        By default, the data will be updated every
        7 days.
        """
        self._update(self._download)
        self._readData()

    def showAvailableData(self):
        """
        Show available data.

        Displays the available column names and associated
        data types, units, and descriptions.
        """
        # Information to be used in output
        fs = ["name", "dtype", "unit", "description"]
        dat = {}
        cols = list(self.vot.columns.values())
        # Collect data
        for i in smo.range(len(cols)):
            dat[i] = {}
            for v in fs:
                dat[i][v] = cols[i].info("attributes", out=None)[v]
        # Determine maximum length of table elements
        maxlens = {v: -1 for v in fs}
        for i in smo.range(len(cols)):
            for v in fs:
                maxlens[v] = max(maxlens[v], len(dat[i][v]))

        # Format the table header
        l = ""
        sep = ""
        for v in fs:
            sep += "=" * maxlens[v] + "  "
            if v == "description":
                # Use left-aligned text for description
                l += ("%-" + str(maxlens[v]) + "s  ") % v
                continue
            l += ("%" + str(maxlens[v]) + "s  ") % v
        # Output header
        print(sep)
        print(l)
        print(sep)

        # Output information
        for i in smo.range(len(cols)):
            l = ""
            for v in fs:
                if v == "description":
                    # Use left-aligned text for description
                    l += ("%-" + str(maxlens[v]) + "s  ") % dat[i][v]
                    continue
                l += ("%" + str(maxlens[v]) + "s  ") % dat[i][v]
            print(l)
        print(sep)

    def getUnitOf(self, col):
        """
        Get unit of column.

        Parameters
        ----------
        col : string
            The name of of column.

        Returns
        -------
        unit : string
            The unit used for this column.
        """
        if not col in self.vot.colnames:
            raise (
                PE.PyAValError(
                    "No such column: " + str(col),
                    solution="Choose one of: " + ", ".join(self.vot.colnames),
                )
            )
        return self.vot[col].info("attributes", out=None)["unit"]

    def _printHRInfo(self, dat):
        """
        Produce human-readable summary of available information.

        Parameters
        ----------
        dat : dictionary
            Output of `selectByPlanetName`.
        """
        # Search keys with/without errors
        kwe, kwoe = [], []
        for k in six.iterkeys(dat):
            if k.find("_error") != -1:
                # Ignore keys containing the '_error' phrase
                continue
            if k.find("_err_") != -1:
                # Ignore keys containing the '_error' phrase
                continue
            if ((k + "_error_min") in dat) or ((k + "_err_min") in dat):
                # Key has error
                kwe.append(k)
            else:
                # Key has no error
                kwoe.append(k)

        # Maximal length of column name
        maxlen = max(list(smo.map(lambda x: len(x), self.getColnames())))
        # Get units
        units = {c: self.getUnitOf(c) for c in self.getColnames()}
        # Maximum length of unit
        mlu = max(list(smo.map(lambda x: len(x), units.values())))

        lines = []

        for k in self.getColnames():
            if (k.find("_error") != -1) or (k.find("_err_") != -1):
                # Ignore keys containing the '_error' phrase
                continue
            if k in kwoe:
                lines.append(
                    ("%" + str(maxlen) + "s") % k
                    + ("  [%" + str(mlu) + "s]  ") % units[k]
                    + str(dat[k])
                )
            else:
                try:
                    # Try _error_ to locate errors
                    ep = "_error_"
                    lines.append(
                        ("%" + str(maxlen) + "s") % k
                        + ("  [%" + str(mlu) + "s]  ") % units[k]
                        + str(dat[k])
                        + "(+"
                        + str(dat[k + ep + "max"])
                        + ", -"
                        + str(dat[k + ep + "min"])
                        + ")"
                    )
                except KeyError:
                    # Try _err_ to locate errors
                    ep = "_err_"
                    lines.append(
                        ("%" + str(maxlen) + "s") % k
                        + ("  [%" + str(mlu) + "s]  ") % units[k]
                        + str(dat[k])
                        + "(+"
                        + str(dat[k + ep + "max"])
                        + ", -"
                        + str(dat[k + ep + "min"])
                        + ")"
                    )
                except:
                    raise

        # Maximum length of output line
        mll = max(list(smo.map(lambda x: len(x), lines)))

        # Print to screen
        print("-" * mll)
        for l in lines:
            print(l)
        print("-" * mll)

    def selectByPlanetName(self, planetName, toScreen=True, caseSensitive=False):
        """
        Get entry by planet name.

        Parameters
        ----------
        planetName : string
            The name of the planet (includes planet letter,
            e.g., "corot-2 b"
        caseSensitive : boolean, optional
            If False (default), the search will be case-insensitive.
        toScreen : boolean, optional
            If True (default), the information on the system is printed
            to screen in human-readable format.

        Returns
        -------
        Data entry : dictionary
            A dictionary with a key for every data column holding
            the associated value from the data table.
        """
        names = [n for n in self.vot["name"]]
        r = pyaC.fuzzyMatch(planetName, names, caseSensitive=caseSensitive, raises=True)
        result = {cn: self.vot[r["index"]][cn] for cn in self.vot.colnames}
        if toScreen:
            self._printHRInfo(result)

        return result

    def getColnames(self):
        """
        Get all column names.

        Returns
        -------
        Columns names : list
            All column names.
        """
        return self.vot.colnames[:]

    def getAllDataAPT(self):
        """
        Get all data as astropy table object.

        Returns
        -------
        votable : astropy.table.table.Table
            All tabulated data as an astropy table
        """
        return copy.copy(self.vot)

    def getAllDataPandas(self):
        """
        Get all data as pandas DataFrame.

        Returns
        -------
        table : DataFrame
            All available data in pandas format.
        """
        if not _ic.check["pandas"]:
            raise (
                PE.PyARequiredImport(
                    "You need to install 'pandas' to use pandas DataFrames.",
                    solution="Install 'pandas' package.",
                )
            )
        return self.vot.to_pandas()

    def __init__(self, skipUpdate=False, forceUpdate=False):
        if not _ic.check["astropy"]:
            raise (
                PE.PyARequiredImport(
                    "The 'astropy' package is not installed. astropy is required to read VO tables.",
                    solution="Please install 'astropy'.",
                )
            )

        configFilename = os.path.join("pyasl", "resBased", "epeuvo.cfg")
        pp.PyAUpdateCycle.__init__(self, configFilename, "ExoUpdate")
        self.dataFileName = os.path.join("pyasl", "resBased", "epeu.vo.gz")
        self._fs = pp.PyAFS()
        if forceUpdate:
            self._update(self._download)
        elif (self.needsUpdate() or (not self._fs.fileExists(self.dataFileName))) and (
            not skipUpdate
        ):
            # Download data if data file does not exist or
            # regular update is indicated
            self._update(self._download)
        self._readData()

    def _download(self):
        """
        Download data.
        """
        self._fs.downloadToFile(
            "http://exoplanet.eu/catalog/votable",
            self.dataFileName,
            clobber=True,
            verbose=False,
            openMethod=gzip.open,
        )



from __future__ import print_function, division
from .idlMod import idlMod
import numpy as np

def cirrange(x, radians=False):
  """
    Force angle into range.
    
    Emulates IDL's `cirrange`.
    
    Parameters
    ----------
    x : float or array
        The angle(s).
    radians : boolean, optional
        If True, angle will be forced into
        0-2pi range. If False (default),
        the 0-360 deg range will be applied.
    
    Returns
    -------
    Modified angle : float or array
        The angle forced into the 0-360 deg
        or 0-2pi radians range.
  """
  if radians:
    m = 2.0 * np.pi
  else:
    m = 360.0
  return x % m
    


from __future__ import print_function, division
import numpy
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo


def turnIntoRebin(CO):
    r"""
    Turn a "normal" fitting object into rebinning fitting object.

    This function accepts a class object representing a model and
    returns another class object extended by the rebinning functionality.

    Parameters
    ----------
    CO : A class object
        The class object describing the model to use
        rebinning.

    Returns
    -------
    Rebinned model : Fitting object
        Another class object extended by the rebinning functionality.
    """

    # Check for evaluate
    if not hasattr(CO, "evaluate"):
        raise (PE.PyANotImplemented("The class object has no evaluate method. Is it really a model?",
                                    where="turnIntoRebin"))
    # Check for other properties and methods
    for p in ["setRebinArray_Ndt", "rebinTimes", "rebinIdent"]:
        if hasattr(CO, p):
            PE.warn(PE.PyANameClash("Class object already has attribute: `" + p + "`. Potentially harmful..",
                                    solution="Check class object, maybe rename attribute", where="turnIntoRebin"))

    class _ModelRebin(CO):
        r"""
        Base class providing rebinning functionality.

        The model is evaluated at more points than actually needed.
        Several points are than averaged to obtain a "binned" model,
        which can, for example, account for finite integration times
        in observations.

        Attributes
        ----------
        rebinTimes : array
            Defined the abscissa values at which
            to evaluate the model to be rebinned afterwards.
        rebinIdent : dict
            A dictionary associating bin number (in the unbinned
            model) with a list holding the bins in `rebinTimes`,
            which are to be averaged to obtain the binned model.
        """

        def __init__(self, *args, **kwargs):
            CO.__init__(self, *args, **kwargs)
            self.rebinTimes = None
            self.rebinIdent = None

        def setRebinArray_Ndt(self, time, N, dt):
            r"""
            Defines the overbinning parameters (`rebinTimes`, `rebinIdent`).

            It is assumed that the time points given in the `time`
            array refer to the center of the time bins and every bin has
            length `dt`. The bins are then subdivided into `N` subintervals;
            the center of each such subinterval becomes
            a point in the overbinned time axis (`rebinTimes`).

            Parameters
            ----------
            time : array
                The time axis of the "observed" (not overbinned)
                transit light-curve.
            N : int
                The number of point into which to subdivide each time
                bin of length `dt`.
            dt : float
                The length of each time bin (on the original not
                oversampled time axis).
            """

            self.rebinTimes = numpy.zeros(time.size * N)
            self.rebinIdent = {}
            for i in smo.range(time.size):
                self.rebinTimes[i * N:(i + 1) * N] = \
                    (time[i] - dt / 2.0) + (numpy.arange(N) * dt) / \
                    float(N) + dt / float(N) / 2.0
                self.rebinIdent[i] = list(range(i * N, (i + 1) * N))

        def evaluate(self, x):
            r"""
            Calculate the model.

            Parameters
            ----------
            x : array
                The abscissa values.

            Returns
            -------
            model : array,
                The binned model.

            Notes
            -----
            This function calculates the model at those time points
            specified by the `rebinTimes` property and saves the result in the
            class property `unbinnedModel`. Then it bins
            according to the definitions in `rebinIdent` and save the resulting model
            in the `binnedModel` property.
            """
            if (self.rebinTimes is None) or (self.rebinIdent is None):
                raise (PE.PyAValError("Rebinning parameters (properties rebinTimes and/or rebinIdent) not appropriately defined.",
                                      solution=["Use setRebinArray_Ndt method.", "Define the properties by accessing them directly."]))
            # Calculate the model using current parameters at the time points
            # defined in the `rebinTimes` array
            self.unbinnedModel = CO.evaluate(self, self.rebinTimes)
            # Build up rebinned model
            self.binnedModel = numpy.zeros(x.size)
            for i in smo.range(x.size):
                self.binnedModel[i] = numpy.mean(
                    self.unbinnedModel[self.rebinIdent[i]])
            # Return the resulting model
            return self.binnedModel

    return _ModelRebin


# The code below is only relevant for Sphinx documentation

class _DummyModel:
    def evaluate(self):
        pass


_ModelRebinDocu = turnIntoRebin(_DummyModel)
_ModelRebinDocu.__name__ = "_ModelRebinDocu"



import numpy as np
import matplotlib.pylab as plt
from PyAstronomy.pyaC import pyaErrors as PE

def sysrem_iter_c(rij, sigij, a, sigij2=None):
    """
    Estimate 'c'
    
    Parameters
    ----------
    rij : 2d array
        Residuals (rij[::,i] = ith of nds observation with nobs data points)
    sigij : 2d array
        Uncertainties
    a : 1d array
        Current estimate of 'a' (length is nobs)
    sigij2 : 2d array, optional
        Square of uncertainties. If given, sigij will be ignores.
    
    Returns
    -------
    c : 1d array
        Estimate of 'c'
    """
    nobs, nds = rij.shape
    
    a2 = a**2
    ss2 = sigij**2 if sigij2 is None else sigij2
    
    # Short form below
    #c = np.zeros(nds)
    #for i in range(nds):
    #    ss = ss2[::,i]
    #    c[i] = np.sum( rij[::,i]*a/ss ) / np.sum( a2/ss )
        
    c = np.sum( ((rij/ss2).T * a).T, axis=0) / np.sum( (a2 * (1/ss2).T).T, axis=0 )
        
    return c

def sysrem_iter_a(rij, sigij, c, sigij2=None):
    """
    Estimate 'a'
    
    Parameters
    ----------
    rij : 2d array
        Residuals (rij[::,i] = ith of nds observation with nobs data points)
    sigij : 2d array
        Uncertainties
    c : 1d array
        Current estimate of 'a' (length is nds)
    sigij2 : 2d array, optional
        Square of uncertainties. If given, sigij will be ignores.
    
    Returns
    -------
    a : 1d array
        Estimate of 'a'
    """
    nobs, nds = rij.shape
    
    c2 = c**2
    ss2 = sigij**2 if sigij2 is None else sigij2
    
    # Short form below
    #a = np.zeros(nobs)
    #for j in range(nobs):
    #    ss = ss2[j,::]
    #    a[j] = np.sum( rij[j,::]*c / ss ) / np.sum( c2/ss )
        
    a = np.sum(rij / ss2 * c, axis=1) / np.sum( c2 * (1/ss2), axis=1 )
    
    return a
    
def sysrem_data_prepare(obs, sigs, ms_obs=True, ms_feat=False):
    """
    Construct data sets for SysRem
    
    Parameters
    ----------
    obs : list of arrays
        Observations
    sigs : list of arrays, optional
        Uncertainties of the observations
    
    Returns
    -------
    rij : 2d array
        The residuals (mean-subtracted observations) so that
        rij[::,i] gives the i-th observation
    sm : 2d array
        The uncertainties in the same format as rij
    a0 : 1d array
        A dummy guess for the 'a' vector (linear between 0 and 1)
    ms_obs : boolean, optional
        Subtract mean from observations (columns of data matrix). Default is True.
    ms_feat : boolean, optional
        Subtract mean from features (e.g., spectral bins). Default is False.
    """
    nds = len(obs)
    nobs = len(obs[0])
    # Residuals and STDs
    rij = np.zeros( (nobs, nds) )
    sm = np.zeros( (nobs, nds) )
    
    if len(obs) != len(sigs):
        raise(ValueError("Obs and sigs must have the same length."))
    
    for n in range(nds):
        if np.isnan(np.min(obs[n])):
            raise(PE.PyAValError(f"NaN value found in observations no. {n}. No NaNs admissible."))
        if np.isnan(np.min(sigs[n])):
            raise(PE.PyAValError(f"NaN value found in error array no. {n}. No NaNs admissible."))
    
    for n in range(nds):
        rij[::,n] = obs[n]
        sm[::,n] = sigs[n]
    
    if ms_obs:
        # Subtract mean for observations/experiments
        for n in range(nds):
            rij[::,n] -= np.mean(rij[::,n])
    if ms_feat:
        # Subtract mean from features/bins
        for n in range(rij.shape[0]):
            rij[n,::] -= np.mean(rij[n,::])

    return rij, sm, np.linspace(0,1,nobs)
    
    
def sysrem_update_rij(rij, a, c):
    """
    Subtract current model from residuals
    
    In matrix notation Tamuz et al. write (Eq. 6) R' = R - c a^T. The operation here
    implemented corresponds to (R' = R - c^T)^T -> R'^T = R^T - a c^T (where ^T is the transpose,
    R is a matrix and a and c are vectors).  
    
    Returns
    -------
    New rij : 2d array
        Updated residuals
    """
    return rij - np.outer(a,c)


class SysRem:
    
    def __init__(self, obs, sigs, ms_obs=True, ms_feat=False, a0=None, ms_warn=True):
        """
        Implementation of the SysRem algorithm.
        
        SysRem was described by
        `Tamuz et al. 2005 (MNRAS 356, 1466) <https://ui.adsabs.harvard.edu/abs/2005MNRAS.356.1466T/abstract>`_
        in the context of correcting systematic effects in samples of
        light curves, but has been applied in other areas such as planetary atmospheric
        spectroscopy.
        
        .. note::
        
            The data (or residuals) are stored such that X[::,0] gives the first observation, i.e.,
            the observation is stored in the first column. The data matrix is, therefore, stored here as
            the transpose of that adopted for the PCA etc..
        
        .. note::
            
            Compared to the presentation by Tamuz et al., the roles of `a` and `c` are exchanged.
        
        Parameters
        ----------
        obs : list of 1d arrays or 2d array
            List of observations (e.g., light curves) or 2d array such that
            obs[::,0] is the first observation. 
        sigs : list of 1d arrays or 2d array
            Uncertainties (same format as obs)
        a0 : 1d array, optional
            Starting values for 'a' parameter (same length as
            the observations). If not given, values linearly
            increasing from 0 to 1 are assumed.
        ms_obs : boolean, optional
            Subtract mean from observations in initial data matrix (columns of data matrix). Default is True.
        ms_feat : boolean, optional
            Subtract mean from features in initial data matrix (e.g., spectral bins). Default is False.
        ms_warn : boolean, optional
            If True (default), a warning will be printed if either ms_obs or ms_feat is True so that the
            data will be manipulated by subtracting mean(s). Set False to suppress warning.
        
        Attributes
        ----------
        rijs : list of 2d arrays
            List of residual arrays. Each call of the `iterate`
            method adds an updated residual array to the list.
        ac : list of tuples of arrays
            The `a` and `c` parameters used to obtain the updated
            residual array from the previous one.
            
        """
        # Convert into list of arrays if 2d array
        if isinstance(obs, np.ndarray) and (obs.ndim == 2):
            obs = [obs[::,i] for i in range(obs.shape[1])]
        if isinstance(sigs, np.ndarray) and (sigs.ndim == 2):
            sigs = [sigs[::,i] for i in range(sigs.shape[1])]
        
        if (ms_obs or ms_feat) and ms_warn:
            PE.warn(PE.PyAValError(f"Be aware: Mean will be subtracted from the data (ms_feat={ms_feat}, ms_obs={ms_obs}).", \
                                   where="SYSREM", \
                                   solution=["Use ms_warn=False to suppress this warning or change ms_feat/obs.", \
                                   "Initial data can be viewed using the attribute rijs[0]."]))
        
        self.rij, self.sm, self.a0 = sysrem_data_prepare(obs, sigs, ms_obs=ms_obs, ms_feat=ms_feat)
        self.sm2 = self.sm**2
        if a0 is not None:
            self.a0 = a0
        
        self.nds = len(obs)
        self.nobs = len(obs[0])
    
        self.rijs = [self.rij]
        self.ac = [(None,None)]
    
    def get_model(self, last=True, a=None, c=None):
        """
        Get model corresponding to a,c vector
        
        Parameters
        ----------
        last : boolean, optional
            If True (default), the model for the last iteration will be
            returned (a,c must not be given then).
        a, c : arrays, optional
            If given, these will be used to calculate the model (`last` must be
            set False)
        """
        
        if (a is not None) and (c is not None) and (not last):
            return np.outer(a, c)
        elif last and (a is None) and (c is None):
            if len(self.rijs) == 0:
                raise(PE.PyAValError("Last model requested, but no model exists yet.", \
                                     where="SysRem", \
                                     solution="Call 'iterate' first."))
            a, c = self.ac[-1][0], self.ac[-1][1]
            return np.outer(a, c)
        else:
            raise(PE.PyAValError("Unknown combination of parameters given.", \
                                 where="SysRem.get_model", \
                                 solution="Specify either a and c or use 'last=True'."))
    
    def get_cumulative_model(self):
        """
        Get the cumulative model (sum of all available models)
        """
        cm = None
        for ac in self.ac:
            if ac[0] is None:
                continue
            m = self.get_model(last=False, a=ac[0], c=ac[1])
            cm = m if cm is None else cm+m
        return cm
    
    def iterate(self, atol=1e-3, rtol=0, imax=1001, a0=None):
        """
        A single SysRem iteration: Remove linear systematic effect
        
        Parameters
        ----------
        atol, rtol : float, optional
            Absolute and relative tolerance for a-c iteration
            required to stop. The pertinent quantity tested is the
            difference between consecutive models divided by the
            uncertainties.
            By default, relative tolerance is ignored.
        imax : int
            Maximum of iterations. Throws exception if convergence
            is not reached earlier.
        a0 : array, optional
            Starting value for SYSREM model PCA-like component. If not
            specified, the value defined during instance initialization
            will be used.
        
        Returns
        -------
        rij : 2d array
            Updated residual matrix with best-fit model subtracted.
        a : array
            Best-fit 'a' values
        c : array
            Best-fit 'c' values
        """
        if a0 is None:
            a = self.a0
        else:
            if a0.shape != self.a0.shape:
                raise(PE.PyAValError("a0 has wrong shape", \
                                     where="SYSREM iterate", \
                                     solution="Needs to be a 1d array with same length as 'observation'."))
            a = a0
        # First ac iteration
        c = sysrem_iter_c(self.rijs[-1], None, a, sigij2=self.sm2)
        a = sysrem_iter_a(self.rijs[-1], None, c, sigij2=self.sm2)
        m = np.outer(a,c)

        converge = False
        for i in range(imax):
            m0 = m.copy()
            c = sysrem_iter_c(self.rijs[-1], None, a, sigij2=self.sm2)
            a = sysrem_iter_a(self.rijs[-1], None, c, sigij2=self.sm2)
            m = np.outer(a,c)
            
            dm = (m-m0)/self.sm
            
            if np.allclose(0,dm,atol=atol,rtol=rtol):
                converge = True
                break

        self.last_ac_iterations = i

        if not converge:
            PE.warn(PE.PyAAlgorithmFailure(f"Iteration (a-c) did not converge with absolute tolerance {atol} and relative tolerance {rtol}. Current value is {np.max(np.abs(dm))}.", \
                                           where="SysRem (iterate)", \
                                           solution="Increase `imax` or adapt atol and/or rtol."))
            
        self.ac.append((a,c))
        self.rijs.append(sysrem_update_rij(self.rijs[-1], a, c))
        return self.rijs[-1], a, c




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Custom objective function
        # ========================
        
        # In[1]:
        
        
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit2 as fuf2
        import scipy.optimize as sco
        import matplotlib
        matplotlib.use('Agg')
        
        np.random.seed(1234)
        
        # Creating a Gaussian with some noise
        # Choose some parameters...
        gPar = {"A":1.0, "sig":10.0, "mu":10.0, "off":1.0, "lin":0.0}
        # Calculate profile
        x = np.arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / np.sqrt(2*np.pi*gPar["sig"]**2) \
            * np.exp(-(x-gPar["mu"])**2/(2*gPar["sig"]**2))
        # Add some noise...
        y += np.random.normal(0.0, 0.002, x.size)
        # ...and save the error bars
        yerr = np.ones_like(x)*0.002
        
        # Create a model object
        gf = fuf2.GaussFit()
        # Set guess values for the parameters
        gf.assignValues({"A":2, "sig":3.77, "off":0.96, "mu":9.5})
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["mu", "sig", "off", "A"])
        
        
        def myobjf(m, pars, x, y, **kwargs):
            """
            Calculate the absolute distance between model and data points and area of the Gaussian
            
            Parameters
            ----------
            m : MBO object
                On setting the objective function later via gf.objf = myobjf, this
                gives access to the parameter values, evaluate method of the model
                and so on.
            pars : list of floats
                Parameter values for which the objective function is to be evaluated.
                A correctly ordered list of free parameter values can be obtained
                by using the method freeParamVals(). The order by names can be inspected
                using the freeParamNames() method.
            x, y : arrays
                Here, this is the data. Note that these variables are specified via the
                'args' parameter later in the call to the optimization algorithm (fmin).
            silent : boolean
                Can be used to suppress output (also specified via 'args' parameter).
            """
            model = m.evaluate(x)
            r = np.sum(np.abs(model - y))
            # Weird dependence on A
            r += 0.6*np.abs(m["A"])
            # Some informative output
            fpv = tuple(zip((m.freeParamNames(), m.freeParamVals())))
            return r
        
        # Show list of free parameters
        print("Names of free parameters: ", gf.freeParamNames())
        
        # Assign objective function. As a result of this assignment,
        # the parameter values of the model will be set to those handed
        # to the objective function prior to call
        gf.addSPLikeObjf(myobjf, "myobjf")
        
        fr = sco.fmin(gf.myobjf, gf.freeParamVals(), args=(x,y))
        print("Fit result: ", fr)
        # Set the parameter values to best-fit
        gf.setFreeParamVals(fr)
        
        gf.parameterSummary()
        # Let us see what we have done...
        plt.plot(x, y, 'bp')
        plt.plot(x, gf.evaluate(x), 'r--')
        plt.show()
        




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Adding two Gaussians
        # ======================
        
        # In[1]:
        
        
        import numpy as np
        import matplotlib.pylab as plt
        # ... and now the funcFit package
        from PyAstronomy import funcFit2 as fuf2
        import scipy.optimize as sco
        import matplotlib
        matplotlib.use('Agg')
        
        np.random.seed(1234)
        
        # Instantiate Gaussian model objects
        gf1 = fuf2.GaussFit()
        gf2 = fuf2.GaussFit()
        # Sum the models (refers to their 'evaluate' methods)
        # Any of +-*/ and ** can be used
        gf = gf1 + gf2
        
        gf.parameterSummary()
        
        gf["A_GF(1)"] = 1
        gf["A_GF(2)"] = 2
        gf["mu_GF(1)"] = 0.0
        gf["mu_GF(2)"] = 3.0
        gf["sig_GF(1)"] = 1.0
        gf["sig_GF(2)"] = 1.0
        
        # Evaluate model and add noise
        x = np.linspace(-4., 6., 200)
        y = gf.evaluate(x) + np.random.normal(0, 0.02, len(x))
        
        # Re-fit model
        # Use filename-like pattern matching to thaw parameters
        gf.thaw(["A_*", "sig_*", "mu_*"])
        fuf2.fitfmin_l_bfgs_b(gf, gf.chisqr, x, y, yerr=0.02)
        
        gf.parameterSummary()
        
        plt.errorbar(x, y, yerr=0.02, fmt="b+", alpha=0.5)
        plt.plot(x, gf.evaluate(x), 'r--', lw=2)
        plt.show()
        



import numpy as np
import os
from PyAstronomy.pyaC import pyaErrors as PE


class Pizzolato2003:

    def _readData(self, fn):
        """
        Read data from file.
        """
        data = np.loadtxt(fn)
        self._tab3 = data[0:7, ::].copy()
        self._tab4 = data[7:, ::].copy()

    def _calcVal(self, tab, row, isat, iprsat, pr):
        """
        Evaluates log10(Lx) or log10(Lx/Lbol) and error

        Parameters
        ----------
        tab : array
            The data table, i.e., Table 3 or 4 from
            Pizzolato et al. 2003.
        isat : int
            Index of value representing saturated level.
        iprsat : int
            Index of limiting rotation period for reaching
            saturation.

        Returns
        -------
        log val : float
            The logarithm (base 10) of Lx or Lx/Lbol
        error log val : float
            The error on that value derived using error
            propagation.
        """
        if pr <= tab[row, iprsat]:
            # Saturated regime
            return tab[row, isat], tab[row, isat+1]
        lnx = tab[row, isat] + 2.0 * \
            np.log10(tab[row, iprsat]) - 2.0*np.log10(pr)
        errlnx = np.sqrt(tab[row, isat+1]**2 +
                         (2.0/(tab[row, iprsat]*np.log(10.0)))**2*tab[row, iprsat+1]**2)
        return lnx, errlnx

    def _findRow(self, tab, val, valname):
        """
        Find corrected row in table.

        Parameters
        ----------
        tab : array
            The data table, i.e., Table 3 or 4 from
            Pizzolato et al. 2003.
        val : float
            Value to be searched for.
        valname : string
            The name of the parameter. Needed in
            case of error.

        Returns
        -------
        row : int
            The index of the correct row.
        """
        row = np.where((val >= tab[::, 0]) & (val < tab[::, 1]))[0]
        if len(row) == 0:
            conc = np.concatenate((tab[::, 0], tab[::, 1]))
            raise (PE.PyAValError(valname + " value of " + str(val) + " is out of range. " +
                                  "Allowed range: " + str(np.min(conc)) + " - " + str(np.max(conc)))
                   )
        return row[0]

    def _checkPr(self, pr):
        """
        Check whether rotation period is valid.
        """
        if pr <= 0.0:
            raise (PE.PyAValError(
                "Rotation period must be >= 0. Given value: " + str(pr)))

    def log10lxbv(self, bv, pr):
        """
        Estimate log10(Lx)

        Parameters
        ----------
        bv : float
            B-V color [mag]
        pr : float
            Rotation period [d]

        Returns
        -------
        log10(Lx) : float
            X-ray luminosity estimate
        Error log10(Lx) : float
            Uncertainty
        """
        self._checkPr(pr)
        row = self._findRow(self._tab4, bv, "B-V")
        return self._calcVal(self._tab4, row, 3, 7, pr)

    def log10lxlbolbv(self, bv, pr):
        """
        Estimate log10(Lx/Lbol)

        Parameters
        ----------
        bv : float
            B-V color [mag]
        pr : float
            Rotation period [d]

        Returns
        -------
        log10(Lx/Lbol) : float
            X-ray luminosity estimate
        Error log10(Lx/Lbol) : float
            Uncertainty
        """
        self._checkPr(pr)
        row = self._findRow(self._tab4, bv, "B-V")
        return self._calcVal(self._tab4, row, 5, 9, pr)

    def log10lxmass(self, m, pr):
        """
        Estimate log10(Lx)

        Parameters
        ----------
        m : float
            Stellar mass [MSun]
        pr : float
            Rotation period [d]

        Returns
        -------
        log10(Lx) : float
            X-ray luminosity estimate
        Error log10(Lx) : float
            Uncertainty
        """
        self._checkPr(pr)
        row = self._findRow(self._tab3, m, "Mass")
        return self._calcVal(self._tab3, row, 3, 7, pr)

    def log10lxlbolmass(self, m, pr):
        """
        Estimate log10(Lx/Lbol)

        Parameters
        ----------
        m : float
            Stellar mass [MSun]
        pr : float
            Rotation period [d]

        Returns
        -------
        log10(Lx/Lbol) : float
            X-ray luminosity estimate
        Error log10(Lx/Lbol) : float
            Uncertainty
        """
        self._checkPr(pr)
        row = self._findRow(self._tab3, m, "Mass")
        return self._calcVal(self._tab3, row, 5, 9, pr)

    def __init__(self):
        r"""
        X-ray luminosity/rotation period relations by Pizzolato et al. 2003.

        Implementation of the relations between X-ray luminosity and
        rotation period given by Pizzolato et al. 2003, A&A 397, 147P.

        To find the X-ray luminosity, the following expression is evaluated

        .. math::
            x = \log_{10} (x_{sat}) + 2\log_{10}(P_{r,sat}) - 2\log_{10}(P_r)

        where x denotes either Lx or Lx/Lbol and 'sat' indicates the value
        at saturation. The coefficients can be found in Tables 3 and 4
        of Pizzolato et al. 2003.
        """
        self._readData(os.path.join(os.path.dirname(
            __file__), "pizzolato2003tables.dat"))



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from . import periodBase
from numpy import var,mean,min,max,pi,sin,cos,sum,arange,zeros,arctan
from PyAstronomy.pyaC import pyaErrors as PE


class LombScargle(periodBase.PeriodBase):
  """
  Calculate the Lomb-Scargle periodogram.
  
  The constructor of *LombScargle* takes a *TimeSeries* instance, i.e., a 
  light curve object, as first argument. It then computes the usual 
  Lomb-Scargle periodogram using a fast algorithm. The frequency array is
  constructed on the fly based on the oversampling keywords, which are
  mandatory in this case. The power is normalized according to the
  prescription of [HB86]_.
    
  The result, i.e., the power, is stored in the class attribute `power`.
    
  .. note::
      Adapted from routine of the same routine in [NR]_ , 
      based on period.pro by Han Wen, August 1996.
    
  Parameters
  ----------
  lc : TimesSeries instance
      The light curve to be analyzed.
  ofac : int
      Oversampling factor.
  hifac : float
      Maximum frequency `freq` = `hifac` * (average Nyquist frequency).
  
  """

  def __calcPeriodogram(self):

    ofac=int(self.ofac); hifac=float(self.hifac)

    if self.freq is None:
      self.__buildFreq(ofac,hifac)
    nout = int(ofac*hifac*len(self.t)/2.)
    vari = var(self.y)
    ave = mean(self.y)

    xdif = max(self.t)-min(self.t)
    xave = 0.5*(max(self.t)+min(self.t))

    # Get starting frequency
    pnow = 1./(xdif*ofac)

    arg = 2.*pi*(self.t-xave)*pnow
    wpr = -2.*sin(0.5*arg)**2
    wpi = sin(arg)
    wr = cos(arg)
    wi = wpi

    py = zeros(nout)
    yy = self.y-ave

    for i in range(nout):
      sumsh = sum(wi*wr)
      sumc = sum((wr-wi)*(wr+wi))
      wtau = 0.5 * arctan(2.*sumsh/sumc)
      swtau = sin(wtau)
      cwtau = cos(wtau)
      ss = wi*cwtau-wr*swtau
      cc = wr*cwtau+wi*swtau
      sums = sum(ss**2)
      sumc = sum(cc**2)
      sumsy = sum(yy*ss)
      sumcy = sum(yy*cc)
      wtemp = wr
      wr   =(wtemp*wpr-wi*wpi)+wr
      wi   =wi*wpr+wtemp*wpi+wi
      py[i]=0.5*(sumcy**2/sumc+sumsy**2/sums)/vari

    self.power = py.copy()
    self.M = 2.*nout/ofac
    self.N = len(self.y)

  def __buildFreq(self,ofac,hifac):
    nout = int(ofac*hifac*len(self.t)/2.)
    xdif = max(self.t)-min(self.t)
    pnow = 1./(xdif*ofac)
    px = pnow + arange(nout)/(ofac*xdif)
    self.freq = px
    return 0

  def __init__(self, lc, ofac, hifac):

    self.freq=None
    self.power=None
    self.ofac=ofac
    self.hifac=hifac
    self.t=lc.time
    self.y=lc.flux
    self.label = {'title': 'Lomb-Scargle periodogram', \
                  'xlabel': 'Frequency', \
                  'ylabel': 'Scargle Power'}
    self.__calcPeriodogram()

  def prob(self, Pn):
    """
    Returns the probability to obtain a power larger than the threshold.
    
    Compute the probability of obtaining a power value as large or larger
    the the threshold based on the noise, which is assumes to be Gaussian.

    .. note::
        *LombScargle* calculates the quantity (N-1)/2.*p=p' (in the formalism of 
        [ZK09]_), which is de facto the normalization
        prescription of [HB86]_. In this
        scheme the probability P(p'>Pn) is given by the following statement:
      
        .. math::
            P(p'>Pn) = \\left(1 - 2 \\frac{Pn}{N-1} \\right)^{(N-3)/2}
  
        If properly normalized to the population variance of the time series, which
        must be known a priori (usually not the case), the
        power :math:`p/p_n=p"` is a direct measure of the SNR as proposed by [Scargle82]_:
        
        .. math::
            P(p">Pn) = exp(-Pn) \\; .
        
        This formula is often used erroneously in this context.
        
      Parameters
      ----------
      Pn : float
          Power threshold
          
      Returns
      -------
      FAP : float
          False alarm probability.
    """
    return (1.-2.*Pn/(self.N-1.))**((self.N-3.)/2.)

  def Prob(self, Pn):
    """
      Outdated -- use "prob" instaed.
    """
    return self.prob(Pn)

  def probInv(self, Prob):
    """
      Returns the minimum power for a given probability level `Prob`.
      
      This is the inverse of `prob(Pn)`. 
      
      Parameters
      ----------
      Prob : float
          Probability
      
      Returns
      -------
      MPL : float
          Minimum power level required.
    """
    return (self.N-1.)/2.*(1.-Prob**(2./(self.N-3.)))

  def ProbInv(self, Prob):
    """
      Outdated -- use "probInv" instaed
    """
    return self.probInv(Prob)


import numpy as np

def lambertPhaseFunction(alpha):
  """
    Calculate phase function for a Lambert sphere.
    
    The phase function of the Lambert sphere is given by:
    
    .. math:: \\Phi(\\alpha) = \\frac{\\sin(\\alpha) + (\\pi - \\alpha)\\cos(\\alpha)}{\\pi} \\; .
    
    Here, :math:`\\alpha` is the phase angle, which is defined as the angle
    between the star and the Earth as seen from the planet. Hence, at phase
    angle zero, the planet is found in opposition. Formally, the phase angle
    can be between 0 and 180 degrees. This function accounts for cases in
    which the given phase angle violates these limits by projecting it back
    into the valid range.
    
    Parameters
    ----------
    alpha : float or array
        The phase angle(s) [deg] at which the phase function
        is to be calculated.
    
    Returns
    -------
    Phase function : float or array
        The values of the phase function at the input
        phase angles.
  """
  a = np.abs(alpha/180.0 * np.pi)
  if not isinstance(a, float):
    a = a - np.floor(a/(2.*np.pi))*(2.*np.pi)
    indi = np.where(a > np.pi)[0]
    a[indi] = 2.*np.pi - a[indi]
    return (np.sin(a) + (np.pi - a)*np.cos(a)) / np.pi
  return float(np.sin(a) + (np.pi - a)*np.cos(a)) / np.pi


from __future__ import print_function, division
from PyAstronomy import pyaC
from PyAstronomy.pyasl import _moduleImportStatus

_modules = ["astroTimeLegacy", "airtovac", "aitoffLegacy", "folding", \
           "keplerOrbit", "inTransit", "statTest", "binning", "smooth", \
           "planet_Teq", "planck", "planetPhase", "fluxConversion", "lineOfSight", \
           "lineOfSight", "dopplerShift", "fluxConversion", "baryvel", "betaSig"]


_modules.extend(["idlMod", "idlCirrange", "rotBroad", "svd", "unred", "estimateSNR", \
                "outlier", "intep", "magnitudes", "eq2hor", "sunpos", "observatory"])


_modules.extend(["moonpos", "moonphase", "localtime", "lineWidth", "cardinalPoint", \
                "twilight", "angularDistance", "airmass", "decimalYear", "broad", \
                "crosscorr", "quadextreme", "readFitsSpec", "posAngle", "sindex", \
                "atomicNo", "coordinates", "writeFitsSpec", "aslExt_1", \
                "gal_uvw", "properMotion"])


from .aslExt_1 import *

for m in _modules:
    _moduleImportStatus[(__name__, m)] = pyaC.pyaimportallfrom(m, "PyAstronomy.pyasl.asl", globals())
 



from . import pyaErrors as PE
import importlib
import traceback


def pyaimportallfrom(mn, pak, globs, cra=False, rf=False, re=False):
    """
    Mimics the behavior of 'from X import *'.

    Get some help here.
    # https://stackoverflow.com/questions/21221358/python-how-to-import-all-methods-and-attributes-from-a-module-dynamically

    Parameters
    ----------
    mn : string
        Module name
    pak : string
        Name of the package to which the imported
        names are added.
    globs : dictionary
        The `globals()` dictionary of that package.
    cra : boolean, optional
        If True, symbol reassignments will be checked
        and some info printed.
    rf : boolean, optional
        If True, a warning will be given on import failure.
        Default is False.
    re : boolean, optional
        If False (default), no exception will be raised on
        import failure (note, also `rf` must be set True to
        make this happen).
        
    Returns
    -------
    status : boolean
        True if import was successful
    exception : string
        If an exception occurred, the text describing the exception. Otherwise
        an empty string
    traceback : list of strings
        If an exception occurred, the traceback. Otherwise an empty string.
    """
    try:
        n = importlib.import_module("." + mn, pak)
        try:
            # If __all__ is defined, use it!
            to_import = n.__all__
        except AttributeError:
            # Otherwise, import all names not starting with an underscore
            to_import = [
                name for name in n.__dict__ if not name.startswith('_')]
            if cra:
                # Check reassignment
                for name in to_import:
                    if name in globs:
                        print("Reassigning: ", name, " from ", n)

        globs.update({name: n.__dict__[name] for name in to_import})
    except Exception as e:
        if rf:
            f = PE.PyAImportFailure("Could not import module " + str(mn) + " to package " + str(pak) + ". " +
                                    "Received message: " + str(e))
            if not re:
                # Report failure without raising exception
                PE.warn(f)
            else:
                # Raise exception
                raise(f)
        return False, str(e), [l.rstrip("\n") for l in traceback.format_exc().splitlines(True)]
    return True, "", ""


class ImportCheck:

    def __init__(self, modules, required=None):
        """
        Checks whether individual modules can be imported.

        Parameters
        ----------
        modules : list of strings
            The names of the modules to be checked.
        required : list of strings, optional
            List of modules which are required. If not found,
            an import error is raised.             
        """
        if required is None:
            required = []
        # List of required modules that could not be imported,
        # list of all modules that could not be imported
        self.requiredFail = []
        self.fail = []

        self.check = {}
        self.versions = {}
        for module in modules:
            self.check[module] = True
            try:
                mi = __import__(module)
            except ImportError:
                self.check[module] = False
                self.fail.append(module)
            except PE.PyARequiredImport:
                self.check[module] = False
                self.fail.append(module)
            except Exception as e:
                # In fact, any error prevents import
                self.check[module] = False
                self.fail.append(module)

            self.versions[module] = "undefined"
            if self.check[module]:
                try:
                    self.versions[module] = mi.__version__
                except:
                    pass

            if (not self.check[module]) and (module in required):
                self.requiredFail.append(module)

        if len(self.requiredFail) > 0:
            ms = ', '.join(self.requiredFail)
            raise(PE.PyARequiredImport(
                "Could not import required module(s): " + ms, solution="Please install " + ms))



import numpy as np
from PyAstronomy import pyasl
from PyAstronomy.pyaC import pyaErrors as PE 

def stringlength_norm(m, norm):
    """
    Normalize string length data set.
    
    Parameters
    ----------
    m : array
        Data array
    norm : string, {default, no}
        If 'default' (default), the data points (mi) will be renormalized according
        to Eq. 3 in Dworetsky 1983, i.e., using mnew = (mi - min(m)) / (2*(max(m) - min(m))).
        If 'no' is specified, the data will not be changed.
    
    Returns
    -------
    ms : array
        The normalized data
    """
    if norm == "default":
        mmin = np.min(m)
        mmax = np.max(m)
        # Dworetsky 1983, Eq. 3
        ms = (m-mmin) / (2 * (mmax-mmin)) - 0.25
    elif norm == "no":
        # Do nothing wrt normalization
        ms = m
    else:
        raise(PE.PyAValError("Unknown value for 'norm': " + str(norm), \
                             where="stringlength_norm", \
                             solution="Use 'default' or 'no'"))
    return ms


def stringlength_pm(p, m, norm="default", closed=True):
    """
    Compute the string length for phased data set.
    
    Parameters
    ----------
    p : array
        The phase array (0-1). Sorted in ascending order.
    m : array
        Data array
    norm : string, {default, no}
        If 'default' (default), the data points (mi) will be renormalized according
        to Eq. 3 in Dworetsky 1983, i.e., using mnew = (mi - min(m)) / (2*(max(m) - min(m))).
        If 'no' is specified, the data will not be changed.
    closed : boolean, optional
        If True (default), first and last point on the phase axis
        will be connected (close the loop).
    
    Returns
    -------
    sl : float
        The string length
    """
    
    if np.any(np.diff(p) < 0):
        raise(PE.PyAValError("Phase array (p) is not sorted in ascending order.", \
                             where="stringlength_pm"))
    
    ms = stringlength_norm(m, norm)

    # String length
    sl = np.sum(np.sqrt( np.diff(p)**2 + np.diff(ms)**2 ))
    if closed:
        # Close the loop ...
        sl += np.sqrt( (p[0] - p[-1] + 1)**2 + (ms[0] - ms[-1])**2 )
    
    return sl


def stringlength_dat(x, m, tps, norm='default', isFreq=False, closed=True):
    """
    Compute string length for data set.
    
    Parameters
    ----------
    x, m : arrays
        x and y coordinates of data points.
    tps : tuple or array
        The trial periods (or frequencies): Either a three-tuple specifying
        (pmin, pmax, nperiods) used by numpy's linspace or an array of trial periods.
        If isFreq is True, tps is interpreted as frequencies.
    isFreq : boolean, optional
        If True, the input tps will be assumed to refer to frequencies instead of periods.
    norm : string, {default, no}
        If 'default' (default), the data points (mi) will be renormalized according
        to Eq. 3 in Dworetsky 1983, i.e., using mnew = (mi - min(m)) / (2*(max(m) - min(m))).
        If 'no' is specified, the data will not be changed.
    closed : boolean, optional
        If True (default), first and last point on the phase axis
        will be connected (close the loop).
    
    Returns
    -------
    Trial periods/frequencies : array
        The tested periods (or frequencies if isFreq is True).
    String length : array
        Associated string lengths
    """
    
    if isinstance(tps, tuple) and len(tps) == 3:
        pers = np.linspace(*tps)
    elif isinstance(tps, np.ndarray):
        pers = tps
    else:
        raise(PE.PyAValError("Unrecognized specification of trial periods or frequencies (tps).", \
                             where="stringlength_dat"))
    
    if isFreq:
        # The input is frequencies. Still need the periods.
        pers = 1.0/pers
    
    ms = stringlength_norm(m, norm)
    
    sls = np.zeros(len(pers))
    for i, p in enumerate(pers):
        phase = pyasl.foldAt(x, p)
        indi = np.argsort(phase)
        sls[i] = stringlength_pm(phase[indi], ms[indi], norm='no', closed=closed)
    
    if isFreq:
        return 1.0/pers, sls
    
    return pers, sls




# -*- coding: utf-8 -*-
from __future__ import print_function, division
from PyAstronomy.pyaC import ImportCheck
from PyAstronomy.pyaC import pyaErrors as PE
import six

_ic = ImportCheck(["numpy", "quantities", "scipy", "matplotlib", "pyfits",
                   "astropy.io.fits", "pandas", "six", "astropy"], required=["numpy", "scipy"])
_moduleImportStatus = {}

from PyAstronomy.pyasl.asl import *
from PyAstronomy.pyasl.resBased import *


def moduleImportStatus(mode="all", tb=False):
    """
    Provide information on model import status.

    Parameters
    ----------
    mode : string, {"all", "fail", "success"}, optional
        If 'all' is given, all modules will be shown. If
        'fail' is given, only failures will be shown and
        the opposite in case of 'success'.
    tb : boolean, optional
        If True, also the traceback of the exception will be
        printed. Default is False.
    """
    mss = ("all", "fail", "success")
    if not mode in mss:
        raise(PE.PyAValError("Unknown mode: " + str(mode),
                             solution="Use any of " + ', '.join(mss)))

    paks = set()
    mods = []
    for k in six.iterkeys(_moduleImportStatus):
        paks.add(k[0])
        mods.append(k[1])
    mlpaks = 0
    mlmods = 0
    for p in paks:
        mlpaks = max(mlpaks, len(p))
    for m in mods:
        mlmods = max(mlmods, len(m))

    lines = []
    mlline = 0
    allk = sorted(list(_moduleImportStatus.keys()))
    for k in allk:
        if mode == "fail":
            if _moduleImportStatus[k]:
                continue
        elif mode == "success":
            if not _moduleImportStatus[k]:
                continue
        lines.append(("%" + str(mlpaks) + "s  " + "%" + str(mlmods) +
                      "s  %5s") % (k[0], k[1], str(_moduleImportStatus[k][0])))
        if not _moduleImportStatus[k][0]:
            lines[-1] += "  " + str(_moduleImportStatus[k][1])
            if tb:
                lines.append("*"*10 + " Start of Traceback " + "*"*10)
                lines.extend(["    "+l for l in _moduleImportStatus[k][2]])
                lines.append("*"*10 + " End of Traceback " + "*"*10)
        mlline = max(mlline, len(lines[-1]))

    sep = "-" * mlline
    head = ("{0: ^" + str(mlline) + "}").format("Import status of 'pyasl'")
    print(sep)
    print(head)
    print(sep)
    for l in lines:
        print(l)
    print(sep)



from PyAstronomy import pyaC
from PyAstronomy.pyasl import _ic, _moduleImportStatus

_modules = ["pizzolato2003", "expCorrRN", "ramirez2005", "stellarAge", "ballesterosBV_T", \
           "abundances", "spectralClass", "mmscetsTable", "roche", "sysrem", "parallactic_angle", \
           "scaleHeight", "solarDirectFlux"]

for m in _modules:
    _moduleImportStatus[(__name__, m)] = pyaC.pyaimportallfrom(m, "PyAstronomy.pyasl.asl.aslExt_1", globals())

if _ic.check["six"]:
    _moduleImportStatus[(__name__, "transCurves")] = pyaC.pyaimportallfrom("transCurves", "PyAstronomy.pyasl.asl.aslExt_1", globals())


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import matplotlib

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
try:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg
except ImportError:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk as NavigationToolbar2TkAgg
from matplotlib.figure import Figure
from PyAstronomy.pyaC import pyaErrors as PE
import numpy
import six.moves.tkinter as tk
import six.moves.tkinter_messagebox as tkMessageBox
import platform


class FFModelPlotFit:
    """
    Plotting and fitting for the model.

    Parameters
    ----------
    x : array
        The abscissa values.
    y : array
        The ordinate values.
    yerr : array, optional
        The error on the points.
    withResiduals : boolean, optional
        If True, the plot will also show the
        residuals. If `yerr` is given, residuals
        will be plotted in units of standard
        deviations. Default is False.
    """

    def __init__(self, x, y, yerr=None, withResiduals=False):
        self.x = x
        self.y = y
        self.yerr = yerr
        self.a = None
        self.withResiduals = withResiduals
        self.modelLine = None
        self.residualLine = None
        self.dataPlot = False
        self.fitLims = None  # range for fit
        self.fitIdx = None  # indices for fit
        self.chi2 = None

    def _setUpAxes(self, f):
        """
        Create axes for plotting
        """
        if not self.withResiduals:
            self.a = f.add_subplot(111)
        else:
            gs = matplotlib.gridspec.GridSpec(3, 1)
            self.a = f.add_subplot(gs[0:2, ::])
            self.ar = f.add_subplot(gs[2, ::], sharex=self.a)

    def plot(self, f, odf):
        """
        Plot the model.

        Parameters
        ----------
        f : matplotlib.Figure
            The figure on which to plot.
        odf : funcFit model class
            The model which is explored.
        """
        # Record whether this is the first call to adjust
        # the autoscale parameter
        firstCall = False
        if self.a is None:
            self._setUpAxes(f)
            firstCall = True

        # Calculate the model
        model = odf.evaluate(self.x)

        if not self.dataPlot:
            # Plot the data
            self.dataPlot = True
            if self.yerr is None:
                self.a.plot(self.x, self.y, 'bp')
                self.a.plot(self.x, self.y,
                            ls='-', color='b', alpha=0.2)
            else:
                self.a.errorbar(self.x, self.y, yerr=self.yerr, fmt='b.')

        if self.modelLine is not None:
            # Remove the last model to show the latest one
            self.modelLine[0].remove()
        # Plot the model
        self.modelLine = self.a.plot(self.x, model, 'r--')

        # Derive chi2 value
        self.chi2 = self.get_chi2(odf)

        if self.withResiduals:
            if self.residualLine is not None:
                self.ar.lines.pop(-1)
            res = self.y - model
            if self.yerr is not None:
                res /= self.yerr
            self.residualLine = self.ar.plot(self.x, res, 'bp')
        if firstCall:
            self.a.autoscale(False)

    def get_chi2(self, odf, lims=None):
        """
        Calculate current chi2 value
        """
        if lims:
            self.fitLims = lims

        r = 0.0
        model = odf.evaluate(self.x)
        if self.fitLims:
            # print "Using lims: ",self.fitLims
            self.fitIdx = numpy.where(numpy.logical_and(
                self.x >= self.fitLims[0], self.x <= self.fitLims[1]))[0]
        else:
            self.fitIdx = numpy.arange(len(self.x))
        if self.yerr is not None:
            r = numpy.sum(
                (self.y[self.fitIdx] - model[self.fitIdx])**2 / self.yerr[self.fitIdx]**2)
        else:
            r = numpy.sum((self.y[self.fitIdx] - model[self.fitIdx])**2)
        return r

    def fit(self, odf, lims=None, verbose=1):
        """
        Fit the model.

        Parameters
        ----------
        odf : funcFit model class
            The model to be fitted.
        """

        if lims:
            if False:  # if old fit ranges -> use old indices... lims == self.fitLims and self.fitIdx:
                odf.fit(self.x[self.fitIdx], self.y[self.fitIdx],
                        yerr=self.yerr[self.fitIdx])
            else:
                self.fitLims = lims
                self.fitIdx = numpy.where(numpy.logical_and(
                    self.x >= self.fitLims[0], self.x <= self.fitLims[1]))[0]
                if verbose > 1:
                    print("fit index: ", self.fitIdx)
                    print("         x ", self.x[self.fitIdx])
                    print("         y ", self.y[self.fitIdx])
                    if self.yerr:
                        print("    yerr ", self.yerr[self.fitIdx])
                    print("     model ", odf.evaluate(self.x[self.fitIdx]))
                if self.yerr is not None:
                    odf.fit(self.x[self.fitIdx], self.y[self.fitIdx],
                            yerr=self.yerr[self.fitIdx])
                else:
                    odf.fit(self.x[self.fitIdx], self.y[self.fitIdx])
        else:
            if self.yerr is not None:
                odf.fit(self.x, self.y, yerr=self.yerr)
            else:
                odf.fit(self.x, self.y)


class SetToDialog(tk.Toplevel):

    def __init__(self, parent, oldVal, pname):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent
        self.newVal = None

        self.inputFrame = tk.Frame(self)
        self.parLabel = tk.Label(
            self.inputFrame, text="Value for parameter: " + str(pname))
        self.parLabel.pack()
        self.inputVal = tk.StringVar()
        self.inputVal.set("% g" % oldVal)
        self.input = tk.Entry(self.inputFrame, textvariable=self.inputVal)
        self.input.pack()
        self.inputFrame.pack(fill=tk.X)

        self.buttonFrame = tk.Frame(self)
        self.okButton = tk.Button(self, text="OK", command=self._okClicked)
        self.caButton = tk.Button(
            self, text="Cancel", command=self._cancelClicked)
        self.okButton.pack(side=tk.RIGHT)
        self.caButton.pack(side=tk.LEFT)

        # Treat return as OK
        self.bind("<Return>", self._okClicked)
        # Treat close as cancel
        self.protocol("WM_DELETE_WINDOW", self._cancelClicked)
        # For a modal dialog
        self.grab_set()

        # Keyboard focus to input entry
        self.input.focus_set()

        self.wait_window(self)

    def _okClicked(self, *args):
        try:
            mf = float(self.inputVal.get())
        except ValueError:
            tkMessageBox.showwarning("Invalid float", "Cannot convert " + self.modEntry.get() + " to float." +
                                     " Make it a valid value to proceed.")
            return
        self.newVal = mf
        self._cancelClicked()

    def _cancelClicked(self, *args):
        self.parent.focus_set()
        self.destroy()


class ShowParameterSummary(tk.Toplevel):

    def __init__(self, parent, odf):
        tk.Toplevel.__init__(self, parent)
        self.wm_title("Parameter Summary")
        self.parent = parent
        self.newVal = None

        self.showFrame = tk.Frame(self)
        self.showText = tk.Text(self.showFrame)
        self.showText.config(background='white')
        self.showText.pack(fill=tk.BOTH, expand=True)
        self.showFrame.pack(fill=tk.BOTH, expand=True)

        self.protocol("WM_DELETE_WINDOW", self._windowClosed)

    def updateInfo(self, odf):
        """
        Update the parameter information.
        """
        self.showText.config(state=tk.NORMAL)
        try:
            self.showText.delete("1.0", tk.END)
        except tk.TclError:
            # Ignore the absence of text
            pass
        lines = odf.parameterSummary(toScreen=False)
        self.showText.config(width=max([len(x) for x in lines]),
                             height=len(lines) + 1)

        t = "".join([x + "\n" for x in lines])
        self.showText.insert(tk.END, t)
        self.showText.config(state=tk.DISABLED)

    def _windowClosed(self):
        """
        The window is closed.
        """
        # Tell the parent that the window does no longer exist
        self.parent.parSumWin = None
        self.destroy()


class ValueSetSummary(tk.Toplevel):
    """
    Inspired from ShowParameterSummary
    """

    def __init__(self, parent, odf):
        tk.Toplevel.__init__(self, parent)
        self.wm_title("Value set code")
        self.parent = parent
        self.newVal = None
        self.odf = odf
        self.showFrame = tk.Frame(self)

        self.textLabel = tk.Label(self.showFrame, text="Model name: ")
        self.textLabel.grid(row=0, column=0)
        self.modelName = tk.StringVar()
        self.modelName.set("model")
        self.modelEntry = tk.Entry(
            self.showFrame, textvariable=self.modelName, width=8)
        self.modelEntry.bind('<Return>', self._updateModelName)
        self.modelEntry.grid(row=0, column=1)

        self.showText = tk.Text(self.showFrame)
        self.showText.config(background='white')
        self.showText.grid(row=1, column=0, columnspan=2)  # .pack(expand=True)

        self.clipboardButton = tk.Button(
            self.showFrame, text="Copy to Clipboard", command=self._copyToClipboard)
        self.clipboardButton.grid(row=2, column=0, columnspan=2)

        self.helpLabel = tk.Label(
            self.showFrame, text="You may use the above code for pasting...")
        self.helpLabel.grid(row=3, column=0, columnspan=2)

        self.showFrame.pack()

        self.protocol("WM_DELETE_WINDOW", self._windowClosed)

    def _updateModelName(self, *args):
        self.updateInfo(self.odf)

    def updateInfo(self, odf):
        """
        Update the parameter information.
        """
        self.showText.config(state=tk.NORMAL)
        try:
            self.showText.delete("1.0", tk.END)
        except tk.TclError:
            # Ignore the absence of text
            pass

        lines = []
        for p in odf.availableParameters():
            lines.append(self.modelName.get() +
                         "[\"" + p + "\"] = " + str(odf[p]))

        self.showText.config(width=max([len(x) for x in lines]),
                             height=len(lines) + 1)
        t = "".join([x + "\n" for x in lines])
        self.showText.insert(tk.END, t)
        self.showText.config(state=tk.DISABLED)

    def _copyToClipboard(self):
        self.clipboard_clear()
        text = self.showText.get("1.0", tk.END)
        self.clipboard_append(text)

    def _windowClosed(self):
        """
        The window is closed.
        """
        # Tell the parent that the window does no longer exist
        self.parent.parSumWin = None
        self.destroy()


class FFModelExplorerList:
    """
    Adapt model interactively.

    Parameters
    ----------
    odf : Instance of OneDFit
        The model to be adapted.
    plotter : Instance of FFModelPlotFit or custom
        Class instance managing the actual plotting.
    """

    def __init__(self, odf, plotter, plist=None):
        self.f = Figure()

        # Save reference to the model
        self.odf = odf
        # Save reference to the plotter
        self.plotter = plotter

        self.leftMask = None
        self.rightMask = None
        self.activeLine = None

        self.root = tk.Tk()
        self.root.wm_title("PyA Model Explorer")
        # Make the widgets expand/shrink as window size changes
        self.root.columnconfigure(0, weight=1)
        self.root.rowconfigure(0, weight=1)

        # Bind the mouse wheel
        if platform.system() == "Linux":
            self.root.bind("<Button-4>", self._mouseWheel)
            self.root.bind("<Button-5>", self._mouseWheel)
        elif platform.system() == "Darwin":
            self.root.bind("<MouseWheel>", self._onWheel)  # for OS X

        # A frame containing the mpl plot
        self.plotFrame = tk.Frame()
        self.plotFrame.pack(fill=tk.BOTH, side=tk.LEFT, expand=True)
        self.canvas = FigureCanvasTkAgg(self.f, master=self.plotFrame)

        # A frame containing the box with selected points
        # and control buttons
        self.controlFrame = tk.Frame(self.root)
        self.controlFrame.pack(side=tk.RIGHT, expand=False, fill=tk.BOTH)
        self.selectedPar = tk.StringVar(self.controlFrame)

        # Get parameters of model
        if plist is None:
            ps = list(self.odf.parameters().keys())
        else:
            ps = plist[:]
        # Set default modification properties
        # Saves these properties for all parameters
        self.modProps = {}
        for p in ps:
            self.modProps[p] = {"modus": "mul",
                                "modValMul": 1.02, "modValAdd": 0.01}

        # Frame containing all parameters
        self.parameterFrame = tk.Frame(
            self.controlFrame, height=2, bd=1, relief=tk.SUNKEN)

        # Dictionaries holding the specific information for each parameter
        self.singleParameterFrames = {}
        self.singleParameterEntry = {}
        self.singleParameterVar = {}
        # knows whether the parameter is free (value=True) or frozen (=False)
        self.singleParameterFree = {}

        # Closures around the functions adapting thaw/freeze
        def frozenChanged(k):
            def change():
                if self.singleParameterFree[k].get() == True:
                    self.odf.thaw(k)
                else:
                    self.odf.freeze(k)
                self._updateDof()
            return change

        # define what happens when a value is set to the entered parameter (closures)
        def hitReturn(k):
            def change(*args):
                self.selectedPar.set(k)
                self.odf[k] = float(self.singleParameterVar[k].get())
                self._parameterValueChanged()
            return change

        # defines what happens when a parameter's value is changed, but not set yet, i.e., not current (closures)
        def parameterValueChanged(k):
            def valueChanged(*args):
                pp, ll = str(self.singleParameterVar[k].get()).find(
                    "."), len(str(self.singleParameterVar[k].get()))
                if round(self.odf[k], ll - pp - 1) != round(self.singleParameterVar[k].get(), ll - pp - 1):
                    self.singleParameterEntry[k].configure(bg="red")
                else:
                    self.singleParameterEntry[k].configure(bg="white")
            return valueChanged

        # Create an entry for each parameter
            # Create a scrollable region
            # Check maximum number of characters for parameter names:
        maxParamLen = 0
        for k in sorted(ps):
            if len(k) > maxParamLen:
                maxParamLen = len(k)
            # Create an entry for each parameter
        for k in sorted(ps):
            x = tk.Frame(self.parameterFrame, height=2,
                         bd=2, relief=tk.SUNKEN, pady=2)
            self.singleParameterFrames[k] = x
            y0 = tk.Radiobutton(x, text=k, variable=self.selectedPar,
                                value=k, width=maxParamLen + 1, indicatoron=0)
            y0.pack(side=tk.LEFT)
            #y1 = tk.StringVar()
            y1 = tk.DoubleVar()
            y1.set(self.odf[k])
            y2 = tk.Entry(x, textvariable=y1, width=8)
            y2.bind('<Return>', hitReturn(k))
            self.singleParameterVar[k] = y1
            self.singleParameterVar[k].trace('w', parameterValueChanged(k))
            self.singleParameterEntry[k] = y2
            y2.pack(side=tk.LEFT)
            self.singleParameterFrames[k].pack()
            modModus = tk.StringVar()
            modModus.set(self.modProps[k]["modus"])
            self.singleParameterFree[k] = tk.BooleanVar()
            self.singleParameterFree[k].set(k in self.odf.freeParamNames())
            y3 = tk.Radiobutton(x, text="Thawed", value=True,
                                variable=self.singleParameterFree[k], command=frozenChanged(k))
            y4 = tk.Radiobutton(x, text="Frozen", value=False,
                                variable=self.singleParameterFree[k], command=frozenChanged(k))
            y3.pack(side=tk.RIGHT)
            y4.pack(side=tk.RIGHT)
            self.parameterFrame.pack(fill=tk.X, expand=False)

        # Set of the menu to select the current parameter
        self.selectedPar.set(ps[0])

        # Chi2 frame:
        # , relief=tk.SUNKEN)
        self.chi2Frame = tk.Frame(self.controlFrame, borderwidth=10)
        self.chi2value = tk.DoubleVar()
        self.chi2value.set(self.plotter.chi2)
        self.dofValue = tk.IntVar()
        # self.dofValue.set(len(self.plotter.fitIdx))
        # self.dofValue.set(100)
        self.dofValue.set(len(self.plotter.x) -
                          len(self.odf.freeParameters()) - 1)
        self.chi2Label = tk.Label(self.chi2Frame, text="Chi2: ")
        self.dofLabel = tk.Label(self.chi2Frame, text="dof: ")
        self.chi2Entry = tk.Entry(
            self.chi2Frame, textvariable=self.chi2value, bd=2, width=10)
        self.dofEntry = tk.Entry(
            self.chi2Frame, textvariable=self.dofValue, bd=2, width=10)
        self.chi2Label.pack(side=tk.LEFT)
        self.chi2Entry.pack(side=tk.LEFT)
        self.dofLabel.pack(side=tk.LEFT)
        self.dofEntry.pack(side=tk.LEFT)
        self.chi2Frame.pack()

        # Frame to bundle mouse-wheel inputs
        self.mouseWheelFrame = tk.Frame(
            self.controlFrame, height=2, bd=3, relief=tk.SUNKEN)
        self.mwmLabel = tk.Label(self.mouseWheelFrame,
                                 text="Mouse wheel manipulation")
        self.mwmLabel.pack()
        # Modify by multiplication or addition (modModus)
        self.modModus = tk.StringVar()
        self.modModus.set("mul")
        # Entry field and radiobutton to specify factor to be used
        self.factorFrame = tk.Frame(self.mouseWheelFrame)
        self.modEntryTextMul = tk.StringVar()
        self.modEntryFactor = tk.Entry(
            self.factorFrame, textvariable=self.modEntryTextMul, width=6)
        self.modEntryFactor.pack(side=tk.LEFT)
        self.radioMultipli = tk.Radiobutton(
            self.factorFrame, text="Multiply", value="mul", variable=self.modModus)
        self.radioMultipli.pack(side=tk.LEFT)
        self.factorFrame.pack(fill=tk.BOTH)
        # Entry field and radiobutton to specify step (delta) to be used
        self.addFrame = tk.Frame(self.mouseWheelFrame)
        self.modEntryTextAdd = tk.StringVar()
        self.modEntryAdd = tk.Entry(
            self.addFrame, textvariable=self.modEntryTextAdd, width=6)
        self.modEntryAdd.pack(side=tk.LEFT)
        self.radioAdd = tk.Radiobutton(
            self.addFrame, text="Add", value="add", variable=self.modModus)
        self.radioAdd.pack(side=tk.LEFT)
        self.addFrame.pack(fill=tk.BOTH)
        # Set text fields for modification factor/step to default
        self.modEntryTextMul.set(
            self.modProps[self.selectedPar.get()]["modValMul"])
        self.modEntryTextAdd.set(
            self.modProps[self.selectedPar.get()]["modValAdd"])
        self.modEntryTextAdd.trace("w", self._modModeChangedAdd)
        self.modEntryTextMul.trace("w", self._modModeChangedMul)
        # Show the frame
#     self.mouseWheelFrame.grid(row=4, column=0, columnspan=3, pady=10)
        self.mouseWheelFrame.pack()

        # React to change in modify Modus
        self.modModus.trace('w', self._modModusChanged)
        # React to a change in the active parameter
        self.selectedPar.trace("w", self._activeParameterChanged)

        dummyLabel = tk.Label(self.controlFrame)
        dummyLabel.pack()

        # Fit button and fit ranges
        self.fitRangeFrame = tk.Frame(
            self.controlFrame, bd=3, relief=tk.SUNKEN)
        self.fit_lo = tk.DoubleVar()
        self.fit_hi = tk.DoubleVar()
        self.fit_lo.set(min(plotter.x))
        self.fit_hi.set(max(plotter.x))
        self.fitRangeLoLim = tk.Entry(
            self.fitRangeFrame, textvariable=self.fit_lo, width=9, bd=2)
        self.fitRangeHiLim = tk.Entry(
            self.fitRangeFrame, textvariable=self.fit_hi, width=9, bd=2)
        self.fitRangeLabel = tk.Label(self.fitRangeFrame, text="Fit range:")
        self.fitButton = tk.Button(
            self.fitRangeFrame, text="Fit", command=self._fitClicked)
        self.fitButton.pack(side=tk.BOTTOM, fill=tk.X)
        self.fitRangeLabel.pack(side=tk.LEFT, fill=tk.X)
        self.fitRangeHiLim.pack(side=tk.RIGHT)
        self.fitRangeLoLim.pack(side=tk.RIGHT)
        self.fitRangeFrame.pack(fill=tk.X)
        #self.fitRangeLoLim.bind('<Return>', self._fitRangeChanged())
        self.fit_lo.trace("w", self._fitRangeChanged)
        self.fit_hi.trace("w", self._fitRangeChanged)
        self.numberClicked = 0
        #self.modModus.trace('w', self._modModusChanged)
        #self.modModus.trace('w', self._modModusChanged)

        dummyLabel = tk.Label(self.controlFrame)
        dummyLabel.pack()

        # , relief=tk.SUNKEN)
        self.showFrame = tk.Frame(self.controlFrame, bd=3)
        self.showFrame.pack(side=tk.TOP)

        self.parSumButton = tk.Button(
            self.showFrame, text="Parameter summary", command=self._parameterSummaryClicked)
        self.parSumButton.pack(side=tk.LEFT)

        self.valSetButton = tk.Button(
            self.showFrame, text="Value set code", command=self._valueSetClicked)
#     self.valSetButton.grid(row=7, column=2)
        self.valSetButton.pack(side=tk.RIGHT)

        # a tk.DrawingArea
#     self.canvas.get_tk_widget().grid(column=0, columnspan=7, row=0, rowspan=10)
        self.canvas.get_tk_widget().pack()
        self.cid = self.f.canvas.mpl_connect(
            'button_press_event', self._mouseButtonClicked)
        self.toolbar = NavigationToolbar2TkAgg(self.canvas, self.plotFrame)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        def _quit():
            # stops main loop
            self.root.quit()
            # this is necessary on Windows to prevent
            # Fatal Python Error: PyEval_RestoreThread: NULL tstate
            self.root.destroy()

        self.quitButton = tk.Button(
            master=self.controlFrame, text='Quit', command=_quit)
        self.quitButton.pack(side=tk.BOTTOM)

        # Plot the model for the first time
        self._parameterValueChanged()

        # Whether or not parameter summary exists
        self.root.parSumWin = None

        if platform.system() == "Darwin":
            self.root.bind("<MouseWheel>", self._onWheel)  # for OS X

    def _parameterActivated(self, *args):
        print(args)

    def _mouseButtonClicked(self, event):
        # print "mouse Button Clicked: ",event
        if event.button != 2:
            return
        # Check whether click occurred on plot axes.
        if event.inaxes is not self.plotter.a:
            # print "not in axis"
            # print event.inaxes, self.plotter.a
            return
        else:
            # print event.xdata, event.ydata
            xrng = self.plotter.a.get_xlim()
            # print
            digits = int(numpy.log10(1. / ((xrng[1] - xrng[0]) / 1000.)) + 2)
            # print digits
            if self.numberClicked == 0:
                self.fit_lo.set(round(event.xdata, digits))
                self.numberClicked = 1
            else:
                self.fit_hi.set(round(event.xdata, digits))
                self.numberClicked = 0
            self.plotter.plot(self.f, self.odf)

            self._fitRangeChanged()

    def _fitRangeChanged(self, *args):
            # print "fitRangeChanged", args
            # print self.fit_lo.get(), self.fit_hi.get()

        if self.leftMask:
            self.plotter.a.collections.remove(self.leftMask)
        if self.rightMask:
            self.plotter.a.collections.remove(self.rightMask)
        if self.activeLine:
            self.activeLine[0].remove()
        try:
            y0, y1 = self.plotter.a.get_ylim()
            y0, y1 = y0 * 10, y1 * 10
            if y0 > 0:
                y0 = 0
            if y1 < 0:
                y1 = 0.
            self.leftMask = self.plotter.a.fill_between(
                [min(self.plotter.x), self.fit_lo.get()], [y0, y0], y2=[y1, y1], alpha=0.2)
            self.rightMask = self.plotter.a.fill_between(
                [self.fit_hi.get(), max(self.plotter.x)], [y0, y0], y2=[y1, y1], alpha=0.2)
        except:
            pass
        # self._parameterValueChanged()
        self.plotter.chi2 = self.plotter.get_chi2(
            self.odf, lims=[self.fit_lo.get(), self.fit_hi.get()])
        self._updateChi2()
        self._updateDof()
        # print self.chi2value.get(), self.fit_lo.get(), self.fit_hi.get()
        self.f.canvas.draw()

    def _updateDof(self):
        self.dofValue.set(len(self.plotter.fitIdx) -
                          len(self.odf.freeParameters()) - 1)

    def _updateChi2(self):
        try:
            if self.plotter.chi2 > 1e-3 and self.plotter.chi2 < 1e4:
                self.chi2value.set("%8.3f" % self.plotter.chi2)
            else:
                self.chi2value.set("%8.3e" % self.plotter.chi2)
        except:
            self.chi2value.set("N/A")

    def _parameterSummaryClicked(self, *args):
        """
        """
        if self.root.parSumWin is None:
            self.root.parSumWin = ShowParameterSummary(self.root, self.odf)
        self.root.parSumWin.updateInfo(self.odf)

    def _valueSetClicked(self, *args):
        """
        """
        if self.root.parSumWin is None:
            self.root.parSumWin = ValueSetSummary(self.root, self.odf)
        self.root.parSumWin.updateInfo(self.odf)

    def _modModusChanged(self, *args):
        """
        Modus for modification changed (multiply or add)
        """
        self.modProps[self.selectedPar.get()]["modus"] = self.modModus.get()

    def _thawFrozenChange(self, *args):
        """
        Called on click to freeze/thaw radiobuttons
        """
        if self.parIsFree.get() == True:
            self.odf.thaw(self.selectedPar.get())
        else:
            self.odf.freeze(self.selectedPar.get())
        self._updateDof()

    def _setToClicked(self):
        """
        Called when "Set to" button is hit
        """
        x = SetToDialog(
            self.root, self.odf[self.selectedPar.get()], self.selectedPar.get())
        if x.newVal is not None:
            self.odf[self.selectedPar.get()] = x.newVal
            self._parameterValueChanged()

    def _fitClicked(self):
        """
        """
        self.plotter.fit(self.odf, lims=[self.fit_lo.get(), self.fit_hi.get()])
        self._parameterValueChanged(allFree=True)

    def _modModeChangedAdd(self, *args):
        self.modProps[self.selectedPar.get(
        )]["modValAdd"] = self.modEntryTextAdd.get()
        self.modProps[self.selectedPar.get()]["modus"] = "add"
        self.modModus.set("add")

    def _modModeChangedMul(self, *args):
        self.modProps[self.selectedPar.get(
        )]["modValMul"] = self.modEntryTextMul.get()
        self.modProps[self.selectedPar.get()]["modus"] = "mul"
        self.modModus.set("mul")

    def _modModeChanged(self, *args):
        """
        Called when mode of modification (add/mul)  is changed
        """
        self.modProps[self.selectedPar.get(
        )]["modValAdd"] = self.modEntryTextAdd.get()
        self.modProps[self.selectedPar.get(
        )]["modValMul"] = self.modEntryTextMul.get()

    def _activeParameterChanged(self, *args):
        """
        Called when the currently active parameter (not its value) is changed.
        """
        # Set the control panels back to that parameter's settings
        pname = self.selectedPar.get()
        newText = "Value:  % g" % (self.odf[pname])
        self.modModus.set(self.modProps[pname]["modus"])
        # Take care of multiply/add radiobuttons
        tmp = self.modProps[pname]["modus"]
        self.modEntryTextAdd.set(self.modProps[pname]["modValAdd"])
        self.modEntryTextMul.set(self.modProps[pname]["modValMul"])
        self.modModus.set(tmp)
        if self.modModus.get() == "mul":
            self.radioMultipli.select()
        else:
            self.radioAdd.select()
        # print

    def _parameterValueChanged(self, allFree=False):
        """
        Called when the value of the current parameter is changed.
        """
        # Update value in label and plot new model
        if not allFree:
            newText = "% g" % (self.odf[self.selectedPar.get()])
            # self.valLabel.config(text=newText)
            # print self.selectedPar.get()
            # print "selected Par: ",self.selectedPar.get()," value: ",self.odf[self.selectedPar.get()]
            self.singleParameterVar[self.selectedPar.get()].set(newText)
        else:
            for p in self.odf.freeParamNames():
                newText = "% g" % (self.odf[p])
                self.singleParameterVar[p].set(newText)

        self.plotter.plot(self.f, self.odf)
        # print self.plotter.chi2
        self._updateChi2()
        self.f.canvas.draw()

    def _onWheel(self, event):
        val = self.odf[self.selectedPar.get()]
        pname = self.selectedPar.get()
        try:
            if self.modModus.get() == "add":
                mf = float(self.modEntryTextAdd.get())
            elif self.modModus.get() == "mul":
                mf = float(self.modEntryTextMul.get())
        except ValueError:
            tkMessageBox.showwarning("Invalid float", "Cannot convert " + self.modEntry.get() + " to float." +
                                     " Make it a valid value to proceed.")
            return
        if self.modModus.get() == "mul":
            if event.delta > 0:
                self.odf[pname] = val * mf
            else:
                self.odf[pname] = val / mf
        else:
            self.odf[pname] = val + mf * event.delta / 2.
        self._parameterValueChanged()

    def _mouseWheel(self, event):
        """
        Mouse wheel moved
        """
        # event.num == 4 -> up
        # event.num == 5 -> down
        # print "mouse wheel  ",event
        val = self.odf[self.selectedPar.get()]
        pname = self.selectedPar.get()
        try:
            if self.modModus.get() == "add":
                mf = float(self.modEntryTextAdd.get())
            elif self.modModus.get() == "mul":
                mf = float(self.modEntryTextMul.get())
        except ValueError:
            tkMessageBox.showwarning("Invalid float", "Cannot convert " + self.modEntry.get() + " to float." +
                                     " Make it a valid value to proceed.")
            return
        if event.num == 4:
            if self.modModus.get() == "mul":
                self.odf[pname] = val * mf
            else:
                self.odf[pname] = val + mf
        elif event.num == 5:
            if self.modModus.get() == "mul":
                self.odf[pname] = val / mf
            else:
                self.odf[pname] = val - mf
        self._parameterValueChanged()

    def show(self):
        """
        Show the GUI.
        """
        try:
            self.canvas.show()
        except AttributeError:
            self.canvas.draw()
        tk.mainloop()



# Set the standard for FFModelExplorer
FFModelExplorer = FFModelExplorerList


def ffmodelExplorer(odf, plotter, version="list", plist=None):
    """
    Instantiate the model explorer.

    Parameters
    ----------
    odf : Instance of OneDFit
        The model to be adapted.
    plotter : Instance of FFModelPlotFit or custom
        Class instance managing the actual plotting.
    version : string, {list}
        The version of model explorer. Currently, only
        'list' is supported.

    Returns
    -------
    ffmod : Model explorer
        An instance of the model explorer.
    """
    if version == "list":
        ffmod = FFModelExplorerList(odf, plotter, plist=plist)
        return ffmod
    elif version == "dropdown":
        PE.warn(PE.PyADeprecationError("Please note that the dropdown version is no longer supported. " +
                                       "The list version will be called instead."))
        ffmod = FFModelExplorerList(odf, plotter, plist=plist)
        return ffmod
    else:
        raise(PE.PyAValError("Unknown version: '" + str(version) + "'",
                             where="ffmodelExplorer",
                             solution="Choose between 'list' and 'dropdown'."))




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Restrictions via algorithm with convenience function
        # ===========================================
        
        # In[1]:
        
        
        from numpy import arange, sqrt, exp, pi, random, ones_like
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit2 as fuf2
        import scipy.optimize as sco
        import matplotlib
        matplotlib.use('Agg')
        
        random.seed(1234)
        
        # Creating a Gaussian with some noise
        # Choose some parameters...
        gPar = {"A":1.0, "sig":10.0, "mu":10.0, "off":1.0, "lin":0.0}
        # Calculate profile
        x = arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / sqrt(2*pi*gPar["sig"]**2) \
            * exp(-(x-gPar["mu"])**2/(2*gPar["sig"]**2))
        # Add some noise...
        y += random.normal(0.0, 0.002, x.size)
        # ...and save the error bars
        yerr = ones_like(x)*0.002
        # Let us see what we have done...
        plt.plot(x, y, 'bp')
        
        # Create a model object
        gf = fuf2.GaussFit()
        
        # Set guess values for the parameters
        gf.assignValues({"A":3, "sig":3.77, "off":0.96, "mu":10.5})
        
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["mu", "sig", "off", "A"])
        
        # Restrict the valid range for sigma
        gf.setRestriction({"sig":[0.01,5]})
        
        # The convenience function 'fitfmin_l_bfgs_b1d' automatically channels
        # the restrictions from the model to the algorithm.
        fuf2.fitfmin_l_bfgs_b(gf, gf.chisqr, x, y, yerr=yerr)
        
        gf.parameterSummary()
        plt.title("Bad fit because of parameter restrictions")
        plt.plot(x, gf.evaluate(x), 'r--')
        plt.show()
        



from __future__ import print_function, division
import unittest


class TestPalTransExampleSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def sanity_calcModel(self):
        """
        Calculate a Pal transit model.
        """
        from PyAstronomy.modelSuite import palTrans
        import matplotlib.pylab as plt
        import numpy as np

        # Create a PalLC instance
        plc = palTrans.PalLC()

        # Set parameter values
        plc["p"] = 0.1  # Planet radius / Stellar radius
        plc["per"] = 1.0  # Orbital period
        plc["a"] = 2.0  # Large semi major axis [R_S]
        plc["i"] = 90.0  # Orbital inclination [deg]
        # Specify limb darkening
        # (quadratic limb-darkening law)
        plc["linLimb"] = 0.4
        plc["quadLimb"] = 0.2
        # Specify T0 (central time of transit)
        plc["T0"] = -0.1
        # Specify binary contribution
        plc["b"] = 0.0

        # Check the parameters
        plc.parameterSummary()

        # Create a time axis
        time = np.arange(1000) / 1000.0 - 0.5

        # Calculate the light curve using above set
        # model parameters
        lightcurve = plc.evaluate(time)

        # Plot the result
        plt.plot(time, lightcurve, "bp")
        # plt.show()

    def sanity_FitModel(self):
        """
        Trying to fit a Pal transit model.
        """
        from PyAstronomy.modelSuite import palTrans
        import matplotlib.pylab as plt
        import numpy as np

        # Create a PalLC instance
        plc = palTrans.PalLC()

        # Set parameter values
        plc["p"] = 0.1  # Planet radius / Stellar radius
        plc["per"] = 1.0  # Orbital period
        plc["a"] = 7.5  # Large semi major axis [R_S]
        plc["i"] = 90.0  # Orbital inclination [deg]
        # Specify limb darkening
        # (quadratic limb-darkening law)
        plc["linLimb"] = 0.4
        plc["quadLimb"] = 0.2
        # Specify T0 (central time of transit)
        plc["T0"] = -0.1
        # Specify binary contribution
        plc["b"] = 0.0

        # Check the parameters
        print("Input parameters: ")
        plc.parameterSummary()

        # Create a time axis
        time = np.arange(100) / 100.0 * 0.2 - 0.2

        # Calculate the light curve using above set
        # model parameters
        lc = plc.evaluate(time)

        # Save the result and add some noise
        flux = lc + np.random.normal(0.0, 0.002, time.size)

        # Now lets try to recover what we put in
        # Choose some "guess" parameters
        plc["p"] = 0.1  # Planet radius / Stellar radius
        plc["per"] = 1.0  # Orbital period
        plc["a"] = 7.5  # Large semi major axis [R_S]
        plc["i"] = 90.0  # Orbital inclination [deg]
        # Specify limb darkening
        # (quadratic limb-darkening law)
        plc["linLimb"] = 0.4
        plc["quadLimb"] = 0.2
        # Specify T0 (central time of transit)
        plc["T0"] = -0.08
        # Specify binary contribution
        plc["b"] = 0.0

        # Assume we want to fit "p", "a", "i", and "T0"
        plc.thaw(["T0", "i"])

        # Before we start fitting, check how the elliptical integrals
        # are evaluated (mpmath or Boost)
        print("Which elliptical integrals are used?: ", plc.whichEllInts())

        # Carry out the fit
        plc.fit(time, flux, yerr=np.ones(time.size) * 0.002)

        print("Fit parameters: ")
        plc.parameterSummary()

        # Plot the result
        plt.plot(time, flux, "bp")
        plt.plot(time, plc.model, "r-")

    #     plt.show()

    def sanity_RebinModel(self):
        """
        Evaluating a rebinned Pal transit model.
        """
        from PyAstronomy.modelSuite import palTrans
        import matplotlib.pylab as plt
        import numpy as np

        # Create a PalLC_Rebin instance
        plc = palTrans.PalLC_Rebin()

        # Set parameter values
        plc["p"] = 0.1  # Planet radius / Stellar radius
        plc["per"] = 1.0  # Orbital period
        plc["a"] = 2.0  # Large semi major axis [R_S]
        plc["i"] = 90.0  # Orbital inclination [deg]
        # Specify limb darkening
        # (quadratic limb-darkening law)
        plc["linLimb"] = 0.8
        plc["quadLimb"] = 0.2
        # Specify T0 (central time of transit)
        plc["T0"] = -0.1
        # Specify binary contribution
        plc["b"] = 0.0

        # Check the parameters
        plc.parameterSummary()

        # Create a time axis
        time = np.arange(50) / 50.0 - 0.51

        # Specify oversampling parameters.
        # Here use 10 points per observed bin.
        plc.setRebinArray_Ndt(time, 10, time[1] - time[0])

        # Calculate the light curve using above set
        # model parameters
        lc = plc.evaluate(time)

        # Plot the result (both the overbinned and final
        # model light-curves)
        plt.plot(plc.rebinTimes, plc.unbinnedModel, "b.-")
        plt.plot(time, plc.binnedModel, "rd--")
        plt.legend(["Overbinned LC", "Averaged LC"])
        # plt.show()



from __future__ import print_function, division
import os
import re
import pickle
import gzip
import numpy as np
from PyAstronomy.pyaC import pyaPermanent as pp
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves.urllib as urllib


class KuruczMT:
    r"""
    Provides access to individual models in a model grid.

    Parameters
    ----------
    fn : string
        Name of the Kurucz model file.

    Attributes
    ----------
    teffs : array
        Available effective temperatures [K] in ascending order.
    loggs : array
        Available loggs in ascending order [cgs].
    models : dictionary
        The key is a tuple of the form: (teff, logg, met)
        and the value is a list of strings representing the
        model.
    met : float
        The logarithmic metallicity of the model grid.

    """

    def _extractModelProperties(self, model):
        r"""
        Extract Teff, logg, and logg from a model.

        Returns
        -------
        Teff, logg, met : float
            Effective temperatures, logg, and log10 of metallicity
            as encoded in the model header.
        """
        r = re.match(r"\s*TEFF\s+([^\s]+)\s+GRAVITY\s+([^\s]+).*", model[0])
        teff = float(r.group(1))
        logg = float(r.group(2))
        r = re.match(r"\s*ABUNDANCE SCALE\s+([^\s]+).*", model[4])
        met = np.log10(float(r.group(1)))
        met = np.floor(met * 10.0 + 0.5) / 10.0
        return teff, logg, met

    def _readModelTable(self, fn):
        r"""
        Read file, extract models, and store in class.

        Parameters
        ----------
        fn : string
            Name of the Kurucz model-grid file. 
        """
        lines = gzip.open(fn, 'rt').readlines()
        if len(lines) == 1:
            lines = lines[0].split("\r")
        models = []
        model = []
        for l in lines:
            if (l.find("TEFF") != -1) and (l.find("GRAVITY") != -1):
                # A new model starts here
                if len(model) > 0:
                    # Add model to list
                    models.append(model[::])
                model = []
            model.append(l.rstrip("\n"))
        if len(model) > 0:
            # Add last model to list (no new TEFF/GRAVITY specification to trigger addition)
            models.append(model[::])
        # Build models dictionary
        teffs = set()
        loggs = set()
        self.met = None
        self.models = {}
        for m in models:
            props = self._extractModelProperties(m)
            self.models[props] = m[::]
            teffs.add(props[0])
            loggs.add(props[1])
            if self.met is None:
                self.met = props[2]
            elif self.met != props[2]:
                raise (PE.PyAValError("Abundance of models differ. This should not happen!\n" +
                                      "The values are: " +
                                      str(props[2]) + " and " + str(self.met),
                                      where="KuruczMT::_readModelTable"))
        # Build array of available models
        self._modelsAvailable = np.zeros(
            (len(teffs), len(loggs)), dtype=bool)
        self.teffs = np.array(sorted(teffs))
        self.loggs = np.array(sorted(loggs))
        for i, teff in enumerate(self.teffs):
            for j, logg in enumerate(self.loggs):
                props = (teff, logg, self.met)
                if props in self.models:
                    self._modelsAvailable[i, j] = True

    def availableTeffs(self):
        r"""
        Get available effective temperatures.

        Returns
        -------
        Teffs : array
            Array of available effective temperatures
            sorted in ascending order. 
        """
        return self.teffs.copy()

    def availableLoggs(self):
        r"""
        Get available logg values.

        Note that not all models are available
        for all logg values.

        Returns
        -------
        Loggs : array
            Array of available loggs sorted in
            ascending order.
        """
        return self.loggs.copy()

    def metallicity(self):
        r"""
        Get metallicity of model grid.

        Returns
        -------
        met : float
            Log10 of the metallicity.
        """
        return self.met

    def modelAvailable(self, teff, logg, met=None):
        r"""
        Determine whether model is available.

        Parameters
        ----------
        teff : float
            Effective temperatures [K]
        logg : float
            Logg [cgs]
        met : float, optional
            Logarithmic metallicity. If not given, the metallicity
            of the model grid is used.

        Returns
        -------
        Availability flag : boolean
            True, if the model is available.
        """
        if met is None:
            met = self.met
        return ((teff, logg, met) in self.models)

    def getModel(self, teff, logg, met=None):
        r"""
        Get a model.

        Parameters
        ----------
        teff : float
            Effective temperatures [K]
        logg : float
            Logg [cgs]
        met : float, optional
            Logarithmic metallicity. If not given, the metallicity
            of the model grid is used.

        Returns
        -------
        Model : list of strings
            The model as found on the file.
        """
        if not self.modelAvailable(teff, logg, met):
            raise (PE.PyAValError(
                ("No model for the combination: Teff = %6.3e, logg = %6.3e, met = " + str(met)) % (teff, logg)))
        if met is None:
            met = self.met
        return self.models[(teff, logg, met)][::]

    def __init__(self, fn):
        if not os.path.isfile(fn):
            raise (PE.PyAValError("No such file: " + str(fn)))
        self._readModelTable(fn)


class KuruczModels:
    r"""
    Provides access to the Kurucz models.
    """

    def getListOfGridsFN(self):
        r"""
        Get filename for list of grids
        """
        return os.path.join("pyasl", "resBased", "kuruczMG.dat.gz")

    def _getListOfModelGrids(self):
        r"""
        Get list of available grids from http://kurucz.harvard.edu/grids.html
        """
        gridFile = self.getListOfGridsFN()
        if not self._fs.fileExists(gridFile):
            # Grid file has to be created
            try:
                with self._fs.requestFile(gridFile, mode='w', openMethod=gzip.open) as f:
                    try:
                        d = urllib.request.urlopen(
                            "http://kurucz.harvard.edu/grids.html").read()
                        d = d.decode("utf8")
                    except urllib.error.URLError as e:
                        raise (PE.PyADownloadError("Could not access URL: http://kurucz.harvard.edu/grids.html\n" +
                                                   "Error: " + str(e),
                                                   solution="Are you online?"))
                    pattern = '<A HREF=\"(.*)\">(.*)</A>'
                    self.grids = {}
                    for grid in re.findall(pattern, d):
                        if grid[0].find("grids") == -1:
                            continue
                        # Find the grid file
                        print("Checking: ", grid[0])
                        try:
                            gfs = urllib.request.urlopen(
                                grid[0]).read().decode("utf8")
                        except urllib.error.HTTPError:
                            # Ignore dead links
                            print("  Ignoring: ", grid[0])
                            continue
                        except urllib.error.URLError as e:
                            print(
                                "  Error opening url: ", grid[0], ". Error message: " + str(e) + ".\n Ignoring...")
                            continue
                        # <img src="/icons/blank.gif" alt="[   ]" width="20" height="20"> <a href="am40ak2odfnew.dat">am40ak2odfnew.dat</a>          13-Apr-2011 15:07  4.2M
                        fns = re.findall('<a href=\"(.*)\">', gfs)
                        # Search for grid files in order of preference
                        # First check for .datcd file
                        gfn = None
                        if gfn is None:
                            # Check for .dat file (with k2)
                            for fn in fns:
                                r = re.match(r"^a.*k2.*\.dat$", fn)
                                if r is not None:
                                    gfn = fn
                                    break
                        if gfn is None:
                            for fn in fns:
                                r = re.match(r"^a.*k2.*\.datcd$", fn)
                                if r is not None:
                                    gfn = fn
                                    break
                        if gfn is None:
                            # Check for .dat file (with any k)
                            for fn in fns:
                                r = re.match(r"^a.*k?.*\.dat$", fn)
                                if r is not None:
                                    gfn = fn
                                    break
                        if gfn is None:
                            # Ignore this grid, there is nothing here
                            continue
                        self.grids[grid[1]] = (grid[0] + gfn)

                    pickle.dump(self.grids, f)
            except:
                # Delete the half-completed file...
                self._fs.removeFile(gridFile)
                raise

        else:
            # File does already exist
            try:
                self.grids = pickle.load(self._fs.requestFile(
                    gridFile, 'r', openMethod=gzip.open))
            except ValueError as ve:
                ffn = self._fs.composeFilename(gridFile)
                per = PE.PyAValError("Reading the pickle file: " + str(ffn) + ", the error: '" + str(ve) + "' occurred.",
                                     solution="Likely, the file was written using Python 3 and you try to read it using Python 2.x. Try to delete the file.")
                raise (per)

    def _abundToStr(self, met):
        r"""
        Convert metallicity into a grid-naming compatible string

        Parameters
        ----------
        met : float
            Log10 of metallicity

        Returns
        -------
        Naming string : string
            E.g., M01, P00 etc.
        """
        if met < 0:
            result = "M"
        else:
            result = "P"
        if (met * 10) - int(abs(met) * 10) > 1e-14:
            raise (PE.PyAValError("Inappropriate met value: " + str(met)))
        met = int(abs(met) * 10)
        if met < 10:
            result += "0"
        result += str(met)
        return result

    def _gridFN(self, name):
        gfn = os.path.join("pyasl", "resBased", name + ".dat.gz")
        return gfn

    def _downloadGrid(self, name):
        r"""
        Download a model-grid file from the Kurucz page.

        Only downloads, if the file does not already exist
        in PyA's data path.

        Parameters
        ----------
        name : string
            The name of the grid.

        Returns
        -------
        gfn : string
            The (relative) name of the (downloaded) file.
        """
        gfn = self._gridFN(name)
        if self._fs.fileExists(gfn):
            return gfn
        print("Downloading model data...")
        print("  Writing data to file: ", gfn, " in PyA data path.")
        self._fs.downloadToFile(self.grids[name], gfn, clobber=True, verbose=False,
                                openMethod=gzip.open)
        return gfn

    def requestModelGrid(self, met, add=""):
        r"""
        Get a model grid.

        Parameters
        ----------
        met : float
            Log10 of the metallicity. For instance use: 0.0, +0.1, +0.5, or -0.5.

        Returns
        -------
        Model grid : KuruczMT
            The model grid enclosed in a class instance, which allows
            easy access to the models.
        """
        if not self.gridAvailable(met, add):
            raise (PE.PyAValError("No appropriate grid available",
                                  solution="Use 'availableGrids' to see what is available."))
        gfn = self._downloadGrid(self._gridName(met, add))
        ffn = self._fs.composeFilename(gfn)
        return KuruczMT(ffn)

    def _gridName(self, met, add):
        r"""
        Compose metallicity and name addition to obtain name of model grid.

        Parameters
        ----------
        met : float
            Log10 of metallicity.
        add : string, optional
            An additional to the name such as "ODFNEW" or "NOVER"
        """
        return "GRID" + self._abundToStr(met) + add

    def gridAvailable(self, met, add=""):
        r"""
        Check whether model grid is available.

        Parameters
        ----------
        met : float
            Log10 of metallicity.

        Returns
        -------
        Availability flag : boolen
            True, if model grid is available.
        """
        return (self._gridName(met, add) in self.grids)

    def availableGrids(self):
        r"""
        All available model grids.

        Returns
        -------
        Available grids : list of strings
            The names of all available model grids.
        """
        return list(self.grids.keys())

    def listGridLinks(self):
        r"""
        List links to available grids
        """
        print("List of grids stored in file:", self.getListOfGridsFN())
        for k, v in self.grids.items():
            print("%20s  " % k, v)

    def __init__(self):
        self._fs = pp.PyAFS()
        self._getListOfModelGrids()


def purgeKuruczData():
    r"""
    Remove downloaded Kurucz data from PyA's data directory
    """
    km = KuruczModels()
    gfn = km.getListOfGridsFN()
    fns = km._fs.globglob(os.path.join("pyasl", "resBased"), "GRID*.dat.gz")
    print("Deleting grid compilation: ", gfn)
    km._fs.removeFile(gfn)
    for fn in fns:
        print("Deleting grid file: ", fn)
        km._fs.removeFile(fn)


def getKuruczModel(teff, logg, met, nameadd=""):
    r"""
    Obtain a Kurucz model

    Parameters
    ----------
    teff : float
        Effective temperature [K]
    logg : float
        Logarithmic surface gravity [cgs]
    met : float
        Logarithmic metallicity, e.g., +0.1.
    nameadd : string, optional
        Name extension of the model grid; for instance,
        "NOVER".

    Returns
    -------
    Model : list of strings
        The requested model.
    """
    km = KuruczModels()
    if not km.gridAvailable(met, nameadd):
        raise (PE.PyAValError("No model grid with logarithmic metallicity of " + str(met) +
                              " and name addition " + str(nameadd) + " available."))
    mt = km.requestModelGrid(met, nameadd)
    if not mt.modelAvailable(teff, logg, met):
        raise (PE.PyAValError("No model available in grid for parameters: teff = %6.3e, logg = %6.3e, met = %6.3e "
                              % (teff, logg, met),
                              solution=["Available Teffs: " + str(list(mt.availableTeffs())),
                                        "Available Loggs: " + str(list(mt.availableLoggs()))]))
    return mt.getModel(teff, logg, met)



import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE

def gyroAgeBarnes(p, bv):
  """
    Calculate gyrochronological age according to Barnes 2007.
    
    The gyrochronological age is calculated according to Eq. 3 in
    Barnes 2007 (ApJ 669, 1167). The derivation of the error follows Eq. 16. 
    
    Parameters
    ----------
    p : float
        Stellar rotation period [d].
    bv : float
        B-V color [mag]. Supported range is 0.4-1.6 mag.
    
    Returns
    -------
    Stellar age : float
        The gyrochronological age [Ga].
    Age error : float
        The error on the age [Ga].
  """
  # See Eq. 3
  n = 0.5189
  a = 0.7725
  b = 0.601
  x = bv-0.4
  
  if bv < 0.4:
    raise(PE.PyAValError("Relation not defined for B-V < 0.4. Value of " + str(bv) + " was given.", \
                         where="gyroAgeBarnes", \
                         solution="Use a value between 0.4 and 1.6 for B-V."))
  if bv > 1.6:
    PE.warn(PE.PyAValError("No calibration for B-V > 1.6. Value given is " + str(bv) + "." +\
                           "The result is an extrapolation.", \
                           where="gyroAgeBarnes"))
  
  # Logarithm of the age [Myr]
  lnt = 1./n * (np.log(p) - np.log(a) - b*np.log(x))
  # Age in Myr
  t = np.exp(lnt)
  # Relative error (Eq. 16)
  dtt = 0.02*np.sqrt(3. + 0.5*lnt**2 + 2.0*p**0.6 + (0.6/x)**2 + (2.4*np.log(x))**2)
  return t/1000., t*dtt/1000.


def chromoAgeRHK(log10RHKprime):
  """
    Calculate the chromospheric age according to Donahue 1998.
    
    Donahue 1998 (ASPC 154, 1235) give a relation between chromospheric
    activity as measured by the R'HK index and the age of late-type stars
    (Eq. 1).
    
    As the level of stellar activity undergoes continuous change, Donahue
    cautions that individual measurements of the activity level yield
    notoriously poor age estimates. As an example, the spread in
    chromospheric solar age resulting from the 11 yr activity cycle is given,
    which amounts to about 2.5 Ga. This gives an idea of the accuracy of
    the estimates.
    
    Parameters
    ----------
    log10RHKprime : float
        Chromospheric activity index log10(R'HK).
    
    Returns
    -------
    Age : float
        Stellar age [Ga].
  """ 
  RHK = 10.0**log10RHKprime
  R5 = 1e5*RHK
  
  logAge = 10.725 - 1.334*R5 + 0.4085*R5**2 - 0.0522*R5**3
  age = 10.0**logAge / 1e9
  return age


import numpy
from PyAstronomy.funcFit import OneDFit

class SinRadVel(OneDFit):
  """
    Spectroscopic radial-velocity (RV) shift due to circular orbital motion.

    *Fit parameters*:
     - P       - float, Orbital period [d]
     - T0      - float, Central transit time [d]
     - K       - float, radial velocity semi-amplitude [km/s]
     - rv0     - float, constant offset in radial velocity [km/s]

   By default all parameters remain frozen.

   The convention is that at phase zero, also the orbital
   radial velocity is zero. With increasing phase the object
   becomes bluer first.
  """

  def __init__(self):
    OneDFit.__init__(self,["P", "T0", "K", "rv0"])
    self.setRootName("SinRV")



  def evaluate(self, x):
    """
      Calculates and returns radial velocity shift according to
      current model parameters.

      Parameters
      ----------
      x : array
          The time stamps at which to calculate the model RV curve.
    """
    y = -numpy.sin(2.0*numpy.pi*(x-self["T0"])/self["P"]) * self["K"] + self["rv0"]
    return y





from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyaC import invertIndexSelection
import numpy as np

def intep(x, y, xinter, boundsError=True, fillValue=None):
  """
    The INTEP interpolation algorithm
    
    The INTEP interpolation algorithm is described
    by Hill 1982, PDAO 16, 67 ("Intep - an Effective
    Interpolation Subroutine"). The implementation
    at hand is based on the FORTRAN code stated
    therein.
    
    The aim of the algorithm is to imitate the curve
    "an experienced scientist" would draw through a
    given set of points. 
    
    Parameters
    ----------
    x : array
        Independent values.
    y : array
        Dependent values.
    xinter : array
        Values at which to interpolate the tabulated
        data given by `x` and `y`.
    boundsError : boolean, optional
        If True, an exception will be raised if values
        need to be extrapolated beyond the limits of
        the given tabulated data. Values beyond the
        limits are simply replaced with the closest
        valid value available, which might not be a
        good approximation. Set this flag to False
        suppress the exception.
    fillValue : float, optional
        If given (i.e., not None), this value will be
        used to represent values outside of the given
        bounds. Note that `boundsError` must be set
        False for this to have an effect. For instance,
        use np.nan.
    
    Returns
    -------
    Interpolated values : array
        Interpolated values at the locations
        specified by `xinter`.
  """
  # Check whether x-array is sorted
  if not np.all((x[1:] - x[0:-1]) > 0.0):
    raise(PE.PyAValError("The array of independent values (`x`) is not sorted in (strictly) " + \
                         "ascending order, but it needs to be.", \
          solution=["Sort the arrays `x` (and `y` accordingly), e.g., using numpy.argsort.", \
                    "Check whether there are duplicate values in `x`."]))
  
  # Create result array
  result = np.zeros(len(xinter))
  # Treat extrapolation points
  ilow = np.where(xinter < min(x))[0]
  if fillValue is None:
    # Use first point beyond limit
    result[ilow] = y[0]
  else:
    result[ilow] = fillValue
  iup = np.where(xinter > max(x))[0]
  if fillValue is None:
    # Use last point beyond limit
    result[iup] = y[-1]
  else:
    result[iup] = fillValue
  
  # No extrapolation (noepo)
  noepo = invertIndexSelection(xinter, np.concatenate((ilow, iup)))
  
  if (len(noepo) < len(xinter)) and boundsError:
    raise(PE.PyAValError("There are " + str(len(xinter) - len(noepo)) + " points, which need to be extrapolated. " + \
          "Attention: Extrapolation is simply done by using the closest valid value.", \
          solution=["Use only interpolation points (`xinter`) within the valid range.",
                    "Set the `boundsError` flag to False (will suppress this exception)."]))
  
  # Loop over all indices of xinter, which were
  # not already subject to extrapolation.
  for i in noepo:
    xp = xinter[i]
    # Index of the first entry in x larger (or equal) than
    # `val`
    infl = np.where(x >= xp)[0][0]
    infl -= 1
    lp1 = 1.0/(x[infl] - x[infl+1])
    lp2 = -lp1
    if infl <= 0:
      # Special treatment for first point
      fp1 = (y[1] -y[0]) / (x[1] - x[0])
    else:
      fp1 = (y[infl+1] - y[infl-1]) / (x[infl+1] - x[infl-1])
    if infl >= (len(x) - 2):
      # Special treatment for last points
      fp2 = (y[-1] - y[-2]) / (x[-1] - x[-2])
    else:
      fp2 = (y[infl+2] - y[infl]) / (x[infl+2] - x[infl])
    
    xpi1 = xp - x[infl+1]
    xpi  = xp - x[infl]
    l1 = xpi1 * lp1
    l2 = xpi * lp2
    result[i] = y[infl]*(1. - 2.*lp1*xpi)*l1**2 + \
                y[infl+1]*(1. - 2.*lp2*xpi1)*l2**2 + \
                fp2*xpi1*l2**2 + fp1*xpi*l1**2
  return result


from __future__ import print_function, division
from PyAstronomy.pyaC import pyaErrors as PE
import re
import six


class ModelNameIdentBase:
    r"""
    Managing the naming of model components.

    This class handles the names of models or model components.
    Individual names or identifiers are composed of a "root name"
    and a "component counter". The root name is supposed to be a
    concise string summarizing the type of model, while the component
    counter is used to distinguish between components
    with the same root name in composed models.

    Parameters
    ----------
    rootName : string, optional
        A concise name for the model (default="").

    Notes
    -----
    The term "specifier" is used to indicate either a string
    containing a variable name or a tuple composed of
    (property, root name, component counter), which
    specifies a variable. In some cases, parts of the specifier
    tuple may be left out.
    """

    def __init__(self, rootName=""):
        self._root = rootName
        self.componentCounter = 0

    def setComponentCounter(self, c):
        self.componentCounter = c

    def getComponentCounter(self):
        return self.componentCounter

    def setRootName(self, root):
        r"""
        Define the root used to build the identifier.

        Parameters
        ----------
        root : string
            Concise description of the model.
        """
        if re.match(r"^[^_^\(^\)]*$", root) is None:
            raise (PE.PyAValError("The proposed root '"+root +
                   "' contains forbidden characters: '_()'."))
        self._root = root

    def identifier(self):
        r"""
        Return the appropriate identifier.

        Returns
        -------
            identifier : string

        Notes
        -----
        The "identifier" reads "root-name(c)",
        where the later part, "(c)", will only be present if the `componentCounter` property
        is larger than zero; in this case, 'c' is that number.
        """
        result = self._root
        if self.componentCounter != 0:
            # Append component counter in parentheses if necessary
            result += "(" + str(self.componentCounter) + ")"
        return result

    def getRoot(self):
        r"""
        Returns the root name.
        """
        return self._root

    def composeVariableName(self, property, rootName=None, counter=None):
        r"""
        Combine property, root name, and counter into a variable name.

        Parameters
        ----------
        property : string
            The property name.
        rootName : string, optional,
            A root name. If None, the instance's own root
            name will be used.
        counter : int, optional,
            The component counter. If None, the instance's
            counter will be used.

        Returns
        -------
        Variable name : string
            The resulting variable name.

        Notes
        -----
        The standard way used to compose variable names is:
        "property_rootname(counter)"
        """
        if rootName is None:
            rootName = self._root
        if counter is None:
            counter = self.componentCounter
        if rootName == "" and counter == 0:
            return property
        result = property + "_" + rootName
        if counter == 0:
            return result
        result += "("+str(counter)+")"
        return result

    def convertSpecifier(self, specifier):
        r"""
        Decompose specifier and compose variable name.

        Parameters
        ----------
        specifier : string or tuple,
            Either a string giving a variable name or a tuple
            specifying property, root name, and counter
            (the later two are optional).

        Returns
        -------
        Decomposed specifier : tuple
            A tuple of (variable name, property, root name,
            component counter).
        """
        component = ""
        counter = 0
        prop = ""
        if isinstance(specifier, tuple):
            prop = specifier[0]
            if len(specifier) >= 2:
                # A component is given
                component = specifier[1]
            if len(specifier) == 3:
                # A counter is also given
                counter = specifier[2]
        else:
            prop = specifier
        return (self.composeVariableName(prop, component, counter), prop, component, counter)

    def specifierToName(self, input):
        r"""
        Convert specifier(s) to variable names.

        Parameter
        ---------
        input : Specifier, list, or dict,
            The input is a single specifier, a list of specifiers, or
            a dictionary in which specifiers are used as keys.

        Returns
        -------
        out : Same type as input,
            Converts all specifiers to variable names. The output is
            the same type as the input.
        """
        if isinstance(input, list):
            # Turn all element into variable names
            result = []
            for el in input:
                result.append(self.convertSpecifier(el)[0])
            return result
        if isinstance(input, dict):
            # Convert all keys to variable names
            result = {}
            for k, v in six.iteritems(input):
                result[self.convertSpecifier(k)[0]] = v
            return result
        # Assume it is a single specifier
        return self.convertSpecifier(input)[0]

    def decomposeVariableName(self, name):
        r"""
        Decomposes variable name into constituents.

        Expects a name of the from "property_root(c)" and returns
        the individual building blocks of the name, i.e., property,
        root name, component counter in that order. If one or more
        of these building blocks is not present, None will
        be returned instead.

        Parameters
        ----------
        name : string
            The variable name.
        """
        r = re.match(r"([^_]+)(_([^\(]*)(\(([0-9]+)\))?)?", name)
        return r.group(1), r.group(3), r.group(5)



# -*- coding: utf-8 -*-
from __future__ import division
from PyAstronomy.funcFit import OneDFit
import numpy as np
from PyAstronomy.modelSuite.XTran import _ZList


class LimBrightTrans(_ZList, OneDFit):
    """
    Planetary transit light-curves for spherical shell model.

    This class implements a model calculating the light curve of a planet
    transiting an optically thin spherical shell of negligible thickness
    (e.g., a stellar chromosphere).

    The model provided by Schlawin et al. 2010 assumes that the thickness
    of the shell is much smaller than the size of the planet.
    The shell is optically thin and thus provides natural limb-brightening.
    The obscured part of the stellar surface is calculated based on computing the
    volume of the intersection of a sphere with a cylinder and then
    taking a partial derivative with respect to the radius of the
    sphere to find its surface area.    

    The code closely follows the IDL procedure located at \
    http://www.astro.washington.edu/agol/.

    *Fit parameters*:
     - `p`     - Rp/Rs (ratio of planetary and stellar radius)
     - `a`     - Semi-major axis of planetary orbit [stellar radii].

     - `per`   - Orbital period [d]
     - `T0`    - Central transit time
     - `i`     - Inclination of orbit [rad]

    By default all parameters remain frozen.
    """

    def __init__(self):
        _ZList.__init__(self, "circular")
        OneDFit.__init__(self, ["p", "a", "i", "T0", "per"])
        self.freeze(["p", "a", "i", "T0", "per"])

        self._zlist = None

    def __ell1(self, k):
        """
        Computes polynomial approximation for the complete elliptic
        integral of the first kind (Hasting's approximation)
        """
        m1 = 1.0 - k**2
        a0 = 1.38629436112
        a1 = 0.09666344259
        a2 = 0.03590092383
        a3 = 0.03742563713
        a4 = 0.01451196212
        b0 = 0.5
        b1 = 0.12498593597
        b2 = 0.06880248576
        b3 = 0.03328355346
        b4 = 0.00441787012
        ek1 = a0 + m1 * (a1 + m1 * (a2 + m1 * (a3 + m1 * a4)))
        ek2 = (b0 + m1 * (b1 + m1 * (b2 + m1 * (b3 + m1 * b4)))) * np.log(m1)
        return ek1 - ek2

    def __ell2(self, k):
        """
        Computes polynomial approximation for the complete elliptic
        integral of the second kind (Hasting's approximation)
        """
        m1 = 1.0 - k**2
        a1 = 0.44325141463
        a2 = 0.06260601220
        a3 = 0.04757383546
        a4 = 0.01736506451
        b1 = 0.24998368310
        b2 = 0.09200180037
        b3 = 0.04069697526
        b4 = 0.00526449639
        ee1 = 1.0 + m1 * (a1 + m1 * (a2 + m1 * (a3 + m1 * a4)))
        ee2 = m1 * (b1 + m1 * (b2 + m1 * (b3 + m1 * b4))) * np.log(1.0 / m1)
        return ee1 + ee2

    def __ell3(self, n, k):
        """
        Computes the complete elliptical integral of the third kind using
        the algorithm of Bulirsch (1965)
        """
        kc = np.sqrt(1.0 - k**2.0)
        p = n + 1.0
        if np.min(p) < 0.0:
            print("Negative p")
        m0 = 1.0
        c = 1.0
        p = np.sqrt(p)
        d = 1.0 / p
        e = kc
        loop = True
        while loop:
            f = c
            c = d / p + f
            g = e / p
            d = (f * g + d) * 2.0
            p = g + p
            g = m0
            m0 = kc + m0
            if np.max(np.abs(1.0 - kc / g)) > 1e-13:
                kc = 2.0 * np.sqrt(e)
                e = kc * m0
            else:
                loop = False
        return 0.5 * np.pi * (c * m0 + d) / (m0 * (m0 + p))

    def evaluate(self, time):
        """
         Calculate a light curve according to the analytical model
         by Schlawin et al. 2010.

         Parameters
         ----------
         time : array
             An array of time points at which the light curve shall be calculated

        Returns
        -------
        Model : array
            The analytical light curve is stored in the property `lightcurve`.

        Notes
        -----

        .. note:: time = 0 -> Planet is exactly in the line of sight (phase = 0).
        """
        self._calcZList(time - self["T0"])

        a = np.zeros(len(self._zlist))
        # Primary transit indices
        itb = np.zeros(len(self._zlist), dtype=bool)
        itb[self._intrans] = 1
        
        indi = np.where((self._zlist + self["p"] < 1.0) & itb)[0]
        
        if len(indi) > 0:
            k = np.sqrt(
                4.0
                * self._zlist[indi]
                * self["p"]
                / (1.0 - (self._zlist[indi] - self["p"]) ** 2)
            )
            a[indi] = (
                4.0
                / np.sqrt(1.0 - (self._zlist[indi] - self["p"]) ** 2)
                * (
                    ((self._zlist[indi] - self["p"]) ** 2 - 1.0) * self.__ell2(k)
                    - (self._zlist[indi] ** 2 - self["p"] ** 2) * self.__ell1(k)
                    + (self._zlist[indi] + self["p"])
                    / (self._zlist[indi] - self["p"])
                    * self.__ell3(
                        4.0
                        * self._zlist[indi]
                        * self["p"]
                        / (self._zlist[indi] - self["p"]) ** 2,
                        k,
                    )
                )
            )

        indi = np.where(
            np.logical_and(
                self._zlist + self["p"] > 1.0, self._zlist - self["p"] < 1.0
            ) & itb
        )[0]
        if len(indi) > 0:
            k = np.sqrt(
                (1.0 - (self._zlist[indi] - self["p"]) ** 2)
                / 4.0
                / self._zlist[indi]
                / self["p"]
            )
            a[indi] = (
                2.0
                / (self._zlist[indi] - self["p"])
                / np.sqrt(self._zlist[indi] * self["p"])
                * (
                    4.0
                    * self._zlist[indi]
                    * self["p"]
                    * (self["p"] - self._zlist[indi])
                    * self.__ell2(k)
                    + (
                        -self._zlist[indi]
                        + 2.0 * self._zlist[indi] ** 2 * self["p"]
                        + self["p"]
                        - 2.0 * self["p"] ** 3
                    )
                    * self.__ell1(k)
                    + (self._zlist[indi] + self["p"])
                    * self.__ell3(-1.0 + 1.0 / (self._zlist[indi] - self["p"]) ** 2, k)
                )
            )

        self.lightcurve = (
            1.0
            - (4.0 * np.pi * ((self["p"] > self._zlist) & itb) * 1.0 + a) / 4.0 / np.pi
        )
        return self.lightcurve



from __future__ import print_function, division
import unittest


class TestConstantsExampleSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_example1(self):
        from PyAstronomy import constants as c

        # Print a summary of available constants
        # on screen
        c.summary()

        # Which unit system is in use?
        print()
        print("Current unit system: ", c.getSystem())

        # Access a constant
        print()
        print("Gravitational constant: ", c.G)
        # The 'f_' prefix is used as a convention. These
        # attributes hold `Quantity` objects as defined in
        # the `quantities` package. These encapsulate value
        # and unit. The prefixless attribute holds only the
        # number.
        print("             with unit: ", c.f_G)
        print("       error with unit: ", c.f_G_err)

        # Change the unit system
        print()
        print("Change the unit system")
        c.setSystem("SI")

        # Which unit system is in use?
        print("Current unit system: ", c.getSystem())

        # Access a constant again...
        print()
        print("Gravitational constant: ", c.G)
        print("             with unit: ", c.f_G)
        print("       error with unit: ", c.f_G_err)

        # Separate value and unit
        print()
        print("Value: ", c.f_G.magnitude, ", units: ", c.f_G.units)

        # Look up details
        print()
        print("What exactly was G?")
        c.constantDetails("G")

        # Apply unit conversion
        print()
        print("Use some other units")
        G_InFeet = c.inUnitsOf("G", "ft**3/(kg * s**2)")
        print("G with feet [ft**3/(kg * s**2)]: ", G_InFeet)

    def testsanity_example2(self):
        from PyAstronomy.constants import PyAConstants

        c = PyAConstants()

        # Which unit system is in use?
        print()
        print("Current unit system: ", c.getSystem())

        # Access a constant
        print()
        print("Gravitational constant: ", c.G)
        # The 'f_' prefix is used as a convention. These
        # attributes hold `Quantity` objects as defined in
        # the `quantities` package. These encapsulate value
        # and unit. The prefixless attribute holds only the
        # number.
        print("             with unit: ", c.f_G)
        print("       error with unit: ", c.f_G_err)




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Applying a relation
        # =====================
        
        # In[1]:
        
        
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit2 as fuf2
        import scipy.optimize as sco
        import matplotlib
        matplotlib.use('Agg')
        
        np.random.seed(1234)
        
        # Creating a Gaussian with some noise
        # Choose some parameters...
        gPar = {"A":1.0, "sig":10.0, "mu":10.0, "off":1.0, "lin":0.0}
        # Calculate profile
        x = np.arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / np.sqrt(2*np.pi*gPar["sig"]**2) \
            * np.exp(-(x-gPar["mu"])**2/(2*gPar["sig"]**2))
        # Add some noise...
        y += np.random.normal(0.0, 0.002, x.size)
        # ...and save the error bars
        yerr = np.ones_like(x)*0.002
        
        # Create a model object
        gf = fuf2.GaussFit()
        
        # Set guess values for the parameters
        gf.assignValues({"A":3, "sig":3.77, "off":0.96, "mu":9.5})
        
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["mu", "sig", "off", "A"])
        
        def relat(sig, off):
            """ Combine values of sig and off """
            return 0.1*sig - off
        
        # 'A' is a function of 'sig' and 'off' (A=f(sig,off)).
        # First parameter is the dependent variable, second is a list of
        # the independent variables, and the third one is theactual
        # functional relation.
        gf.relate("A", ["sig", "off"], func=relat)
        
        fr = sco.fmin(gf.chisqr, gf.freeParamVals(), args=(x,y))
        print("Fit result: ", fr)
        # Set the parameter values to best-fit
        gf.setFreeParamVals(fr)
        
        gf.parameterSummary()
        
        # Let us see what we have done...
        plt.plot(x, y, 'bp')
        plt.plot(x, gf.evaluate(x), 'r--')
        plt.show()
        
